import os
import tkinter as tk
from tkinter import filedialog, ttk, messagebox

from openbiolink.graph_creation import graphCreationConfig as gcConst
from openbiolink.gui import gui as gui


class SplitFrame(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller
        self.info_text = ""
        titles_panel = tk.Frame(self)
        self.info = tk.Button(titles_panel, text=" help ", command=lambda: gui.show_info_box(self.info_text))
        self.title = tk.Label(titles_panel, text="(2) Split Creation", font=controller.title_font)

        self.mode_panel = self._create_mode_element(self)

        self.option_panel = tk.Frame(self)
        self.left_box = tk.LabelFrame(self.option_panel, text='graph files')
        self.file_edge_el = self._create_edge_path_el(self.left_box)
        self.file_tn_el = self._create_tn_path_el(self.left_box)
        self.file_nodes_el = self._create_nodes_path_el(self.left_box)
        self.right_box = tk.Frame(self.option_panel)
        self.right_rand_box = tk.LabelFrame(self.right_box, text='parameter for random split')
        self.rand_box_el = self._create_rand_options_el(self.right_rand_box)
        self.right_time_box = tk.LabelFrame(self.right_box, text='parameter for time slice')
        self.time_box_el = self._create_time_options_el(self.right_time_box)

        self.buttons_panel = tk.Frame(self)
        self.next_button = tk.Button(self.buttons_panel, text="Next", command=lambda: self.next_page(), height=1,
                                     width=15)
        self.prev_button = tk.Button(self.buttons_panel, text="Back",
                                     command=lambda: self.controller.show_previous_frame(), height=1, width=15)

        # packing
        titles_panel.pack(side="top", fill="x", pady=10)
        self.title.pack(side='left', pady=10, padx=15)
        self.info.pack(side="right", fill="x", pady=5, padx=15)
        self.mode_panel.pack(side='top', fill='both', padx=15, expand=True)
        self.option_panel.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        self.left_box.pack(side='left', fill='both', expand=True)
        self.file_edge_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        self.file_tn_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        self.file_nodes_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        self.right_box.pack(side='left', fill='both', expand=True)
        self.right_rand_box.pack(fill='both', expand=True, padx=(10, 0))
        self.rand_box_el.pack(fill='both', expand=True)

        ttk.Separator(self, orient='horizontal').pack(side='top', fill='x', pady=(15, 0), padx=10, anchor='s')
        self.buttons_panel.pack(side='bottom', padx=15, fill='x')
        self.prev_button.pack(side='left', anchor='w', pady=(5, 10))
        self.next_button.pack(side='right', anchor='e', pady=(5, 10))

    def _create_mode_element(self, parent):
        el = tk.LabelFrame(parent, text="Split Mode")
        self.mode = tk.StringVar(value='rand')
        random_box = tk.Radiobutton(el, text='random split', variable=self.mode, value='rand',
                                    command=self.toggl_tts_elements)
        time_box = tk.Radiobutton(el, text='time slice', variable=self.mode, value='time',
                                  command=self.toggl_tts_elements)
        random_box.pack(side='left', padx=5, anchor='w')
        time_box.pack(side='left', padx=5, anchor='w')
        return el

    def toggl_tts_elements(self):
        if self.mode.get() == 'time':
            self.right_rand_box.pack_forget()
            self.rand_box_el.pack_forget()
            self.right_time_box.pack(fill='both', expand=True, padx=(10, 0))
            self.time_box_el.pack(fill='both', expand=True)
        elif self.mode.get() == 'rand':
            self.right_rand_box.pack(fill='both', expand=True, padx=(10, 0))
            self.rand_box_el.pack(fill='both', expand=True)
            self.right_time_box.pack_forget()
            self.time_box_el.pack_forget()

    def _create_time_options_el(self, parent):
        el = tk.Frame(parent)
        file_edge_el = self._create_tmo_edge_path_el(el)
        file_tn_el = self._create_tmo_tn_path_el(el)
        file_nodes_el = self._create_tmo_nodes_path_el(el)
        file_edge_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        file_tn_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        file_nodes_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        return el

    def _create_rand_options_el(self, parent):
        el = tk.Frame(parent)

        # test frac
        self.test_frac = tk.StringVar(value='0.2')
        test_frac_frame = tk.Frame(el)
        test_frac_label = tk.Label(test_frac_frame, text='test set fraction:')
        test_frac_value = tk.Entry(test_frac_frame, textvariable=self.test_frac, width=5)
        test_frac_info = tk.Label(test_frac_frame, text='as float, e.g. 0.2', font=self.controller.info_font)

        # cross val
        self.crossval = tk.BooleanVar(value=False)
        crossval_box = tk.Checkbutton(el, text='cross validation', variable=self.crossval)
        self.folds = tk.StringVar(value='5')
        folds_frame = tk.Frame(el)
        folds_label = tk.Label(folds_frame, text='folds / validation set fraction:')
        folds_value = tk.Entry(folds_frame, textvariable=self.folds, width=5)
        folds_info = tk.Label(folds_frame, text='folds as int\n validation fraction as float',
                              font=self.controller.info_font)

        # packing
        test_frac_frame.pack(side='top', padx=5, pady=20, anchor='w')
        test_frac_label.pack(side='left', padx=5, anchor='w')
        test_frac_value.pack(side='left', anchor='w')
        test_frac_info.pack(side='left', anchor='w')
        ttk.Separator(el, orient='horizontal').pack(side='top', fill='x', pady=5, padx=5, anchor='w')
        crossval_box.pack(side='top', padx=5, anchor='w')
        folds_frame.pack(side='top', padx=5, anchor='w')
        folds_label.pack(side='left', padx=5, anchor='w')
        folds_value.pack(side='left', anchor='w')
        folds_info.pack(side='left', anchor='w')
        return el

    # todo style: 6 times same function
    def _create_edge_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.edge_path = tk.StringVar()
        edge_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_edge_file())
        edge_label = tk.Entry(el, textvariable=self.edge_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="edge path:").pack(side='left', anchor='w', padx=5, pady=5)
        edge_button.pack(side='right', anchor='w', padx=5, pady=5)
        edge_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_tn_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.tn_path = tk.StringVar()
        tn_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_tn_file())
        tn_label = tk.Entry(el, textvariable=self.tn_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="true negative edge path:").pack(side='left', anchor='w', padx=5, pady=5)
        tn_button.pack(side='right', padx=5, pady=5)
        tn_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_nodes_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.nodes_path = tk.StringVar()
        nodes_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_nodes_file())
        nodes_label = tk.Entry(el, textvariable=self.nodes_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="nodes file path").pack(side='left', anchor='w', padx=5, pady=5)
        nodes_button.pack(side='right', anchor='w', padx=5, pady=5)
        nodes_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_tmo_edge_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.tmo_edge_path = tk.StringVar()
        edge_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_tmo_edge_file())
        edge_label = tk.Entry(el, textvariable=self.tmo_edge_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="edge path:").pack(side='left', anchor='w', padx=5, pady=5)
        edge_button.pack(side='right', anchor='w', padx=5, pady=5)
        edge_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_tmo_tn_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.tmo_tn_path = tk.StringVar()
        tn_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_tmo_tn_file())
        tn_label = tk.Entry(el, textvariable=self.tmo_tn_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="true negative edge path:").pack(side='left', anchor='w', padx=5, pady=5)
        tn_button.pack(side='right', padx=5, pady=5)
        tn_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_tmo_nodes_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.tmo_nodes_path = tk.StringVar()
        nodes_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_tmo_nodes_file())
        nodes_label = tk.Entry(el, textvariable=self.tmo_nodes_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="nodes file path").pack(side='left', anchor='w', padx=5, pady=5)
        nodes_button.pack(side='right', anchor='w', padx=5, pady=5)
        nodes_label.pack(fill='x', padx=5, pady=5)
        return el

    def browse_edge_file(self):
        self.edge_path.set(filedialog.askopenfilename())

    def browse_tn_file(self):
        self.tn_path.set(filedialog.askopenfilename())

    def browse_nodes_file(self):
        self.nodes_path.set(filedialog.askopenfilename())

    def browse_tmo_edge_file(self):
        self.tmo_edge_path.set(filedialog.askopenfilename())

    def browse_tmo_tn_file(self):
        self.tmo_tn_path.set(filedialog.askopenfilename())

    def browse_tmo_nodes_file(self):
        self.tmo_nodes_path.set(filedialog.askopenfilename())

    def update(self):
        # fixme check if graph creation action is performed
        graph_files_folder = os.path.join(self.controller.ARGS_LIST_GLOBAL[1], gcConst.GRAPH_FILES_FOLDER_NAME)
        edge_path = os.path.join(graph_files_folder, 'edges.csv')
        if os.path.exists(edge_path) or 'GraphCreationFrame' in self.controller.selected_frames:
            self.edge_path.set(edge_path)
        tn_edge_path = os.path.join(graph_files_folder, 'TN_edges.csv')
        if os.path.exists(tn_edge_path) or 'GraphCreationFrame' in self.controller.selected_frames:
            self.tn_path.set(tn_edge_path)
        nodes_path = os.path.join(graph_files_folder, 'ALL_nodes.csv')
        if os.path.exists(nodes_path) or 'GraphCreationFrame' in self.controller.selected_frames:
            self.nodes_path.set(nodes_path)

    def next_page(self):
        self.controller.ARGS_LIST_TRAIN_TEST_SPLIT = []
        self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.append('-s')
        self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--edges', self.edge_path.get()])
        self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--tn_edges', self.tn_path.get()])
        self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--nodes', self.nodes_path.get()])

        if self.mode.get() == 'time':
            if (self.tmo_edge_path.get() == "") or (self.tmo_tn_path.get() == '') or (self.tmo_nodes_path.get() == ""):
                messagebox.showerror('ERROR', 'Please provide all three t-minus one paths.')
                return
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--mode', 'time'])
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--tmo_edges', self.tmo_edge_path.get()])
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--tmo_tn_edges', self.tmo_tn_path.get()])
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--tmo_nodes', self.tmo_nodes_path.get()])

        elif self.mode.get() == 'rand':
            if self.test_frac.get() == '':
                messagebox.showerror('ERROR', 'Please provide a test set fraction.')
                return
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--mode', 'rand'])
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--test_frac', self.test_frac.get()])

            if self.crossval.get():
                self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.append('--crossval')  # todo crossval
                self.controller.ARGS_LIST_TRAIN_TEST_SPLIT.extend(['--val', self.folds.get()])

        self.controller.show_next_frame()



import json
import os

import pandas

import openbiolink.evaluation.evalConfig as evalConst
from openbiolink import globalConfig as globConst


def write_mappings(node_label_to_id=None, node_types_to_id=None, relation_label_to_id=None):
    model_dir = os.path.join(os.path.join(globConst.WORKING_DIR, evalConst.EVAL_OUTPUT_FOLDER_NAME),
                             evalConst.MODEL_DIR)
    os.makedirs(model_dir, exist_ok=True)

    if node_label_to_id is not None:
        with open(os.path.join(model_dir, evalConst.MODEL_ENTITY_NAME_MAPPING_NAME), 'w') as file:
            json.dump(node_label_to_id, file, indent=4, sort_keys=True)
    if node_types_to_id is not None:
        with open(os.path.join(model_dir, evalConst.MODEL_ENTITY_TYPE_MAPPING_NAME), 'w') as file:
            json.dump(node_types_to_id, file, indent=4, sort_keys=True)
    if relation_label_to_id is not None:
        with open(os.path.join(model_dir, evalConst.MODEL_RELATION_TYPE_MAPPING_NAME), 'w') as file:
            json.dump(relation_label_to_id, file, indent=4, sort_keys=True)


def write_metric_results(metrics_results):
    eval_dir = os.path.join(globConst.WORKING_DIR, evalConst.EVAL_OUTPUT_FOLDER_NAME)
    os.makedirs(eval_dir, exist_ok=True)

    with open(os.path.join(eval_dir, evalConst.EVAL_OUTPUT_FILE_NAME), 'w') as fp:
        json_metrics = {x.value: y for x, y in metrics_results.items()}
        json.dump(json_metrics, fp, indent=4)


def read_mapping(path):
    with open(path) as f:
        string = f.read()
        return json.loads(string)


def read_corrupted_triples(path, sep='\t'):
    heads_path = os.path.join(path, evalConst.CORRUPTED_HEADS_FILE_NAME)
    tails_path = os.path.join(path, evalConst.CORRUPTED_TAILS_FILE_NAME)
    if os.path.exists(heads_path) and os.path.exists(tails_path):
        col_names = [evalConst.CORRUPTED_GROUP_COL_NAME,
                     globConst.NODE1_ID_COL_NAME,
                     globConst.EDGE_TYPE_COL_NAME,
                     globConst.NODE2_ID_COL_NAME,
                     globConst.VALUE_COL_NAME]
        all_corrupted_heads = pandas.read_csv(heads_path, names=col_names, sep=sep)
        all_corrupted_tails = pandas.read_csv(tails_path, names=col_names, sep=sep)
        head_groups = all_corrupted_heads[evalConst.CORRUPTED_GROUP_COL_NAME].unique()
        tail_groups = all_corrupted_tails[evalConst.CORRUPTED_GROUP_COL_NAME].unique()
        corrupted_head_dict = {}
        corrupted_tail_dict = {}
        for group in head_groups:
            df = all_corrupted_heads[all_corrupted_heads[evalConst.CORRUPTED_GROUP_COL_NAME] == group]
            df.drop(evalConst.CORRUPTED_GROUP_COL_NAME, axis=1)
            df.reset_index(drop=True, inplace=True)
            h, r, t, _ = tuple(df.iloc[0].values)
            corrupted_head_dict[(h, r, t)] = df
        for group in tail_groups:
            df = all_corrupted_tails[all_corrupted_tails[evalConst.CORRUPTED_GROUP_COL_NAME] == group]
            df.drop(evalConst.CORRUPTED_GROUP_COL_NAME, axis=1)
            df.reset_index(drop=True, inplace=True)
            h, r, t, _ = tuple(df.iloc[0].values)
            corrupted_tail_dict[(h, r, t)] = df
        return corrupted_head_dict, corrupted_tail_dict



from openbiolink.graph_creation.file_processor.mapping.mapDisGeNetProcessor import MapDisGeNetProcessor
from openbiolink.graph_creation.file_processor.mapping.mapDrugCentralPubchemProcessor import MapDrugCentralPubchemProcessor
from openbiolink.graph_creation.file_processor.mapping.mapStringProcessor import MapStringProcessor
from openbiolink.graph_creation.file_processor.mapping.mapUniEnsNcbiProcessor import MapUniEnsNcbiProcessor
from openbiolink.graph_creation.file_processor.mapping.mapUniUniNcbiProcessor import MapUniUniNcbiProcessor



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeStringAction(DbMetadataEdge):
    NAME = 'Edge - STRING - Gene Gene (Action)'

    # current link can be extracted from https://stringdb-static.org/cgi/download.pl?sessionId=FuXJ9a0fkSMB&species_text=Homo+sapiens
    URL = "https://stringdb-static.org/download/protein.actions.v11.0/9606.protein.actions.v11.0.txt.gz"
    OFILE_NAME = "STRING_gene_gene_actions.tsv.gz"  # tab separated txt file
    COLS = ['item_id_a', 'item_id_b', 'mode', 'action', 'is_directional', 'a_is_acting', 'score']
    FILTER_COLS = ['item_id_a', 'item_id_b', 'mode', 'action', 'is_directional', 'a_is_acting', 'score']
    HEADER = 1
    DB_TYPE = DbType.DB_EDGE_STRING_ACTION

    def __init__(self):
        super().__init__(url=self.URL,
                         ofile_name=self.OFILE_NAME,
                         dbType=self.DB_TYPE)



import logging
import math
import random
import sys

import numpy
import pandas

import openbiolink.graphProperties as graphProp
from openbiolink import globalConfig
from openbiolink import globalConfig as glob
from openbiolink import utils
from openbiolink.graph_creation.metadata_edge import edgeMetadata as meta
from openbiolink.train_test_set_creation.sampler import NegativeSampler
from openbiolink.train_test_set_creation.trainTestSetWriter import TrainTestSetWriter

random.seed(glob.RANDOM_STATE)
numpy.random.seed(glob.RANDOM_STATE)


class TrainTestSetCreation():
    """
    Manager class for handling the creation of train test splits given a graph

        Attributes
                ----------
                    all_nodes : pandas.DataFrame
                        DataFrame with all nodes, columns = globalConfig.COL_NAMES_NODES
                    all_tp : pandas.DataFrame
                        DataFrame with edges from the positive graph, i.e. all positive examples
                        columns = globalConfig.COL_NAMES_EDGES + globalConfig.VALUE_COL_NAME
                    tp_edgeTypes : [str]
                        list of all edge types present in the positive examples
                    all_tn : pandas.DataFrame
                        DataFrame with edges from the negative graph, i.e. all negative examples
                        columns = globalConfig.COL_NAMES_EDGES + globalConfig.VALUE_COL_NAME
                    tn_edgeTypes : [str]
                        list of all edge types present in the negative examples
                    meta_edges_dic : {str: (str, str, str)}
                        dictionary for all possible h,r,t combinations, mapped to their types. The key consists of
                        %s_%s_%s'%(node1Type,edgeType,node2Type) the value (node1Type, edgeType, node2Type)
                    tmo_nodes : pandas.DataFrame
                        DataFrame with all nodes present in the t-1 graph, columns = globalConfig.COL_NAMES_NODES
                    tmo_all_tp : pandas.DataFrame
                        DataFrame with edges from the positive t-1 graph, i.e. all positive t-1 examples
                        columns = globalConfig.COL_NAMES_EDGES + globalConfig.VALUE_COL_NAME
                    tmo_tp_edgeTypes : [str]
                        list of all edge types present in the positive t-1 examples
                    tmo_all_tn : pandas.DataFrame
                        DataFrame with edges from the negative t-1 graph, i.e. all negative t-1 examples
                        columns = globalConfig.COL_NAMES_EDGES + globalConfig.VALUE_COL_NAME
                    tmo_tn_edgeTypes : [str]
                        list of all edge types present in the negative t-1 examples

    """

    def __init__(self,
                 graph_path,
                 tn_graph_path,
                 all_nodes_path,
                 sep='\t',
                 # meta_edge_triples=None, #nicetohave (1) split for subsample of edges, define own meta edges
                 t_minus_one_graph_path=None,
                 t_minus_one_tn_graph_path=None,
                 t_minus_one_nodes_path=None):

        self.writer = TrainTestSetWriter()
        with open(all_nodes_path) as file:
            self.all_nodes = pandas.read_csv(file, sep=sep, names=globalConfig.COL_NAMES_NODES)
            self.all_nodes = self.all_nodes.sort_values(by=globalConfig.COL_NAMES_NODES).reset_index(drop=True)

        with open(graph_path) as file:
            self.all_tp = pandas.read_csv(file, sep=sep, names=globalConfig.COL_NAMES_EDGES)
            self.all_tp[globalConfig.VALUE_COL_NAME] = 1
            self.all_tp = self.all_tp.sort_values(by=globalConfig.COL_NAMES_EDGES).reset_index(drop=True)
        self.tp_edgeTypes = list(self.all_tp[globalConfig.EDGE_TYPE_COL_NAME].unique())

        with open(tn_graph_path) as file:
            self.all_tn = pandas.read_csv(file, sep=sep, names=globalConfig.COL_NAMES_EDGES)
            self.all_tn[globalConfig.VALUE_COL_NAME] = 0
            self.all_tn = self.all_tn.sort_values(by=globalConfig.COL_NAMES_EDGES).reset_index(drop=True)

        self.tn_edgeTypes = list(self.all_tn[globalConfig.EDGE_TYPE_COL_NAME].unique())

        self.meta_edges_dic = {}

        for metaEdge in utils.get_leaf_subclasses(meta.EdgeMetadata):
            edgeType = str(metaEdge.EDGE_INMETA_CLASS.EDGE_TYPE)
            node1Type = str(metaEdge.EDGE_INMETA_CLASS.NODE1_TYPE)
            node2Type = str(metaEdge.EDGE_INMETA_CLASS.NODE2_TYPE)
            if edgeType in self.tp_edgeTypes:
                self.meta_edges_dic['%s_%s_%s' % (node1Type, edgeType, node2Type)] = (node1Type, edgeType, node2Type)

        # nicetohave (2) check for transient onto edges
        # transitiv_IS_A_edges = utils.check_for_transitive_edges(self.all_tp[self.all_tp[ttsConst.EDGE_TYPE_COL_NAME] == 'IS_A'])
        # transitiv_PART_OF_edges = utils.check_for_transitive_edges(self.all_tp[self.all_tp[ttsConst.EDGE_TYPE_COL_NAME] == 'PART_OF'])
        # if transitiv_IS_A_edges:
        #    print('WARNING: transient edges in IS_A: ({a},b,c) for a IS_A b and a IS_A c', transitiv_IS_A_edges)
        # if transitiv_PART_OF_edges:
        #    print('WARNING: transient edges in PART_OF: ({a},b,c) for a PART_OF b and a PART_OF c',
        #          transitiv_PART_OF_edges)

        # for time slices
        if not (bool(t_minus_one_graph_path) == bool(t_minus_one_tn_graph_path) == (bool(t_minus_one_nodes_path))):
            logging.error('either all three or none of these variables must be provided')
            sys.exit()
        if t_minus_one_nodes_path and t_minus_one_graph_path and t_minus_one_tn_graph_path:
            with open(t_minus_one_nodes_path) as file:
                self.tmo_nodes = pandas.read_csv(file, sep=sep, names=globalConfig.COL_NAMES_NODES)
                self.tmo_nodes = self.tmo_nodes.sort_values(by=globalConfig.COL_NAMES_NODES).reset_index(drop=True)

            with open(t_minus_one_graph_path) as file:
                self.tmo_all_tp = pandas.read_csv(file, sep=sep, names=globalConfig.COL_NAMES_EDGES)
                self.tmo_all_tp[globalConfig.VALUE_COL_NAME] = 1
                self.tmo_all_tp = self.tmo_all_tp.sort_values(by=globalConfig.COL_NAMES_EDGES).reset_index(drop=True)
            self.tmo_tp_edgeTypes = list(self.all_tp[globalConfig.EDGE_TYPE_COL_NAME].unique())

            with open(t_minus_one_tn_graph_path) as file:
                self.tmo_all_tn = pandas.read_csv(file, sep=sep, names=globalConfig.COL_NAMES_EDGES)
                self.tmo_all_tn[globalConfig.VALUE_COL_NAME] = 0
                self.tmo_all_tn = self.tmo_all_tn.sort_values(by=globalConfig.COL_NAMES_EDGES).reset_index(drop=True)
            self.tmo_tn_edgeTypes = list(self.all_tp[globalConfig.EDGE_TYPE_COL_NAME].unique())

    def random_edge_split(self, test_frac=None, val=None, crossval=None):
        if not val:
            val = 0.2
        if not test_frac:
            test_frac = 0.2

        # create positive and negative examples
        positive_samples = self.all_tp.copy()
        negative_sampler = NegativeSampler(self.meta_edges_dic, self.tn_edgeTypes, self.all_tn.copy(), self.all_nodes)
        negative_samples = negative_sampler.generate_random_neg_samples(positive_samples)
        all_samples = (positive_samples.append(negative_samples, ignore_index=True)).reset_index(drop=True)
        all_samples = utils.remove_inconsistent_edges(all_samples).reset_index(drop=True)

        # generate, train-, test-, validation-sets
        test_set = all_samples.sample(frac=test_frac, random_state=glob.RANDOM_STATE)
        train_val_set = all_samples.drop(list(test_set.index.values))
        test_set = utils.remove_parent_duplicates_and_reverses(remain_set=test_set, remove_set=train_val_set)

        nodes_in_train_val_set = train_val_set[globalConfig.NODE1_ID_COL_NAME].tolist() \
                                 + train_val_set[globalConfig.NODE2_ID_COL_NAME].tolist()
        new_test_nodes = self.get_additional_nodes(old_nodes_list=nodes_in_train_val_set,
                                                   new_nodes_list=self.all_nodes[
                                                       globalConfig.ID_NODE_COL_NAME].tolist())
        if new_test_nodes:
            logging.info(
                'The test set contains nodes, that are not present in the trainings-set. These edges will be dropped.')  # nicetohave (6): option to keep edges with new nodes
            test_set = self.remove_edges_with_nodes(test_set, new_test_nodes)
        nodes_in_test_set = test_set[globalConfig.NODE1_ID_COL_NAME].tolist() \
                            + test_set[globalConfig.NODE2_ID_COL_NAME].tolist()
        if graphProp.DIRECTED:
            train_val_set = utils.remove_reverse_edges(remain_set=train_val_set, remove_set=test_set)

        if crossval:
            train_val_set_tuples = self.create_cross_val(train_val_set, val)
            new_val_nodes = None
            for i, train_val_set_tuple in enumerate(train_val_set_tuples):
                train_set, val_set = train_val_set_tuple
                new_val_nodes = self.get_additional_nodes(
                    old_nodes_list=train_set[globalConfig.NODE1_ID_COL_NAME].tolist()
                                   + train_set[globalConfig.NODE2_ID_COL_NAME].tolist(),
                    new_nodes_list=nodes_in_train_val_set)
                if new_val_nodes:  # nicetohave (6)
                    logging.info(
                        'Validation set %d contains nodes, that are not present in the trainings-set. These edges will be dropped.' % i)
                    val_set = self.remove_edges_with_nodes(val_set, new_val_nodes)
                    train_val_set_tuples[i] = (train_set, val_set)

        else:
            train_val_set_tuples = [(train_val_set, pandas.DataFrame())]
            new_val_nodes = None
        if graphProp.DIRECTED:
            train_val_set_tuples = [(utils.remove_reverse_edges(remain_set=t, remove_set=v), v)
                                    for t, v in train_val_set_tuples]
        train_val_set_tuples = [(t, utils.remove_parent_duplicates_and_reverses(remain_set=v, remove_set=t))
                                for t, v in train_val_set_tuples]

        self.writer.print_sets(train_val_set_tuples=train_val_set_tuples,
                               new_val_nodes=new_val_nodes,
                               test_set=test_set,
                               new_test_nodes=new_test_nodes,
                               nodes_in_train_val_set=set(nodes_in_train_val_set),
                               nodes_in_test_set=set(nodes_in_test_set))
        # nicetohave (3) option to remove examples with new nodes
        return train_val_set_tuples, test_set

    def time_slice_split(self):

        # nicetohave (4) like that, neg samples are restricted to edge_types appearing in test_sample --> good idea?
        # nicetohave (4) idea: calculate nodes like above, then tmo_nodes= test_nodes --> mehr auswahl bei neg examples
        tmo_positive_samples = self.tmo_all_tp
        tmo_negative_sampler = NegativeSampler(self.meta_edges_dic,
                                               self.tmo_tn_edgeTypes,
                                               self.tmo_all_tn,
                                               self.tmo_nodes)
        tmo_negative_samples = tmo_negative_sampler.generate_random_neg_samples(tmo_positive_samples)
        # todo remove not consistent edges
        tmo_negative_samples[globalConfig.VALUE_COL_NAME] = 0
        tmo_all_samples = (tmo_positive_samples.append(tmo_negative_samples, ignore_index=True)).reset_index(
            drop=True)  # todo ist append nicht in pace?
        train_set = tmo_all_samples
        test_positive_samples, vanished_positive_samples = utils.get_diff(self.all_tp, self.tmo_all_tp,
                                                                          ignore_qscore=True)
        test_tn_samples, vanished_tn_samples = utils.get_diff(self.all_tn, self.tmo_all_tn, ignore_qscore=True)
        if not vanished_positive_samples.empty or not vanished_tn_samples.empty:
            logging.info('Some edges existing in the first time slice are no longer present in the second one')
            self.writer.print_vanished_edges(vanished_positive_samples.append(vanished_tn_samples))
        test_negative_sampler = NegativeSampler(self.meta_edges_dic,
                                                self.tn_edgeTypes,
                                                test_tn_samples,
                                                self.all_nodes)
        test_negative_samples = test_negative_sampler.generate_random_neg_samples(test_positive_samples)
        test_negative_samples[globalConfig.VALUE_COL_NAME] = 0

        test_set = (test_positive_samples.append(test_negative_samples, ignore_index=True)).reset_index(drop=True)
        test_set = utils.remove_parent_duplicates_and_reverses(remain_set=test_set, remove_set=train_set)
        new_test_nodes = self.get_additional_nodes(
            old_nodes_list=self.tmo_nodes[globalConfig.ID_NODE_COL_NAME].tolist(),
            new_nodes_list=self.all_nodes[globalConfig.ID_NODE_COL_NAME].tolist())
        if new_test_nodes:
            logging.info(
                'The test set contains nodes, that are not present in the trainings-set. These edges will be removed.')  # nicetohave (6)
            test_set = self.remove_edges_with_nodes(test_set, new_test_nodes)

        if graphProp.DIRECTED:
            train_set = utils.remove_reverse_edges(remain_set=train_set, remove_set=test_set)

        train_val_set_tuples = [(train_set, pandas.DataFrame())]
        nodes_in_train_set = train_set[globalConfig.NODE1_ID_COL_NAME].tolist() \
                             + train_set[globalConfig.NODE2_ID_COL_NAME].tolist()
        nodes_in_test_set = test_set[globalConfig.NODE1_ID_COL_NAME].tolist() \
                            + test_set[globalConfig.NODE2_ID_COL_NAME].tolist()
        self.writer.print_sets(train_val_set_tuples=train_val_set_tuples,
                               test_set=test_set,
                               nodes_in_train_val_set=nodes_in_train_set,
                               nodes_in_test_set=nodes_in_test_set,
                               new_test_nodes=new_test_nodes)
        return (train_val_set_tuples, test_set)

        # niceToHave (5) slice size?
        # extra nur für diff? --> eig nciht --> eig schon weil nur pos

    @staticmethod
    def remove_edges_with_nodes(samples: pandas.DataFrame, nodes: set):
        """
        :param samples: pandas.DataFrame containing edges of the samples, node col names according to globalConfig
        :param nodes: list of new nodes
        :return: returns samples pandas.DataFrame with all edges, that contain a node from new_nodes, removed
        """
        new_node_in_edge = samples[globalConfig.NODE1_ID_COL_NAME].isin(nodes) \
                           | samples[globalConfig.NODE2_ID_COL_NAME].isin(nodes)
        edges_with_new_nodes = samples.loc[new_node_in_edge]
        return samples.drop(list(edges_with_new_nodes.index.values))

    @staticmethod
    def get_additional_nodes(old_nodes_list: list, new_nodes_list: list):
        """

        :param old_nodes_list:
        :param new_nodes_list:
        :return: returns a set containing all nodes that are in new_nodes_list but not in old_nodes_list
        """
        old_set = set(old_nodes_list)
        new_set = set(new_nodes_list)
        return new_set - old_set

    @staticmethod
    def create_cross_val(df: pandas.DataFrame, n_folds):
        nel_total, _ = df.shape
        if n_folds == 0 or n_folds == 1 or (n_folds > 1 and not float(n_folds).is_integer()):
            logging.error("provided folds are not possible!")
            raise Exception(
                "fold entry must be either an int>1 (number of folds) or a float >0 and <1 (validation fraction)")

        if n_folds < 1:  # n_folds is fraction
            n_folds = math.ceil(1 / n_folds)
        n_folds = int(n_folds)
        nel_per_chunk = math.ceil(nel_total / n_folds)
        rand_index = list(df.index.values)
        random.shuffle(rand_index)

        bounds = []
        for i in range(n_folds):
            bounds.append(i * nel_per_chunk)
        bounds.append(nel_total)

        chunks = []
        for i in range(n_folds):
            chunks.append(rand_index[bounds[i]: bounds[i + 1]])

        folds_indices = []
        for i in range(n_folds):
            train_chunk_indices = [(x + i) % n_folds for x in range(n_folds - 1)]
            val_chunk_index = (n_folds - 1 + i) % n_folds
            train_indices = [element for chunk_index in train_chunk_indices for element in chunks[chunk_index]]
            val_indices = chunks[val_chunk_index]
            folds_indices.append((train_indices, val_indices))

        folds = []
        for train_indices, val_indices in folds_indices:
            folds.append((df.loc[train_indices], df.loc[val_indices]))

        return folds



from openbiolink.edgeType import EdgeType
from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.nodeType import NodeType


class InMetaEdgeDrugCentral(InfileMetadata):
    CSV_NAME = "DB_DrugCentral_drug_indication_dis.csv"

    USE_COLS = [
        "struct_id",
        "umls_cui",
    ]
    NODE1_COL = 1
    NODE2_COL = 0
    QSCORE_COL = None
    NODE1_TYPE = NodeType.DIS
    NODE2_TYPE = NodeType.DRUG
    EDGE_TYPE = EdgeType.DIS_DRUG
    INFILE_TYPE = InfileType.IN_EDGE_DRUGCENTRAL_IND
    MAPPING_SEP = None

    def __init__(self):
        super().__init__(csv_name=InMetaEdgeDrugCentral.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaEdgeDrugCentral.INFILE_TYPE)



from openbiolink.graph_creation.file_processor.edge import *
from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.file_processor.mapping import *
from openbiolink.graph_creation.file_processor.onto import *
from openbiolink.graph_creation.file_processor.onto_mapping import *



from openbiolink.edgeType import EdgeType
from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.nodeType import NodeType


class InMetaEdgeStringActivation(InfileMetadata):
    CSV_NAME = "DB_STRING_gene_activation_gene.csv"
    USE_COLS = ['item_id_a', 'item_id_b', 'score']
    NODE1_COL = 0
    NODE2_COL = 1
    QSCORE_COL = 2
    NODE1_TYPE = NodeType.GENE
    NODE2_TYPE = NodeType.GENE
    EDGE_TYPE = EdgeType.GENE_ACTIVATION_GENE
    INFILE_TYPE = InfileType.IN_EDGE_STRING_ACTIVATION

    MAPPING_SEP = None

    def __init__(self):
        super().__init__(csv_name=self.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=self.INFILE_TYPE)



import os

import pandas

import openbiolink.train_test_set_creation.ttsConfig as ttsConst
from openbiolink import globalConfig as globConst


class TrainTestSetWriter():
    def __init__(self):
        self.folder_path = os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME)
        os.makedirs(self.folder_path, exist_ok=True)

    def print_sets(self, train_val_set_tuples: list, test_set, nodes_in_train_val_set, nodes_in_test_set,
                   new_val_nodes=None, new_test_nodes=None):
        if new_test_nodes is None:
            new_test_nodes = []
        if new_val_nodes is None:
            new_val_nodes = [None] * len(train_val_set_tuples)

        num_folds = len(train_val_set_tuples)

        test_set[globConst.COL_NAMES_SAMPLES].to_csv(os.path.join(self.folder_path, ttsConst.TEST_FILE_NAME),
                                                     sep='\t',
                                                     index=False,
                                                     header=False)

        train_val_nodes_df = pandas.DataFrame({'id': list(nodes_in_train_val_set)})
        train_val_nodes_df['nodeType'] = [x[0] for x in train_val_nodes_df['id'].str.split('_')]
        train_val_nodes_df.to_csv(os.path.join(self.folder_path, ttsConst.TRAIN_VAL_NODES_FILE_NAME),
                                  sep='\t',
                                  index=False,
                                  header=False)
        test_nodes_df = pandas.DataFrame({'id': list(nodes_in_test_set)})
        test_nodes_df['nodeType'] = [x[0] for x in test_nodes_df['id'].str.split('_')]
        test_nodes_df.to_csv(os.path.join(self.folder_path, ttsConst.TEST_NODES_FILE_NAME),
                             sep='\t',
                             index=False,
                             header=False)

        if new_test_nodes:
            with open(os.path.join(self.folder_path, ttsConst.NEW_TEST_NODES_FILE_NAME), 'w', newline='\n') as file:
                file.writelines(list('\n'.join(new_test_nodes)))

        if num_folds > 1:
            t_set, v_set = train_val_set_tuples[0]
            fill_train_set = t_set.append(v_set)
            fill_train_set[globConst.COL_NAMES_SAMPLES].to_csv(os.path.join(self.folder_path, ttsConst.TRAIN_FILE_NAME),
                                                               sep='\t',
                                                               index=False,
                                                               header=False)
            self.folder_path = os.path.join(self.folder_path, ttsConst.CROSS_VAL_FOLDER_NAME)

        i = 0
        for (train_set, val_set), new_val_nodes_for_fold in zip(train_val_set_tuples, new_val_nodes):
            if num_folds > 1:
                fold_folder_path = os.path.join(self.folder_path, ttsConst.FOLD_FOLDER_PREFIX + str(i))
            else:
                fold_folder_path = self.folder_path
            os.makedirs(fold_folder_path, exist_ok=True)
            if not train_set.empty:
                train_set[globConst.COL_NAMES_SAMPLES].to_csv(os.path.join(fold_folder_path, ttsConst.TRAIN_FILE_NAME),
                                                              sep='\t',
                                                              index=False,
                                                              header=False)
            if not val_set.empty:
                val_set[globConst.COL_NAMES_SAMPLES].to_csv(os.path.join(fold_folder_path, ttsConst.VAL_FILE_NAME),
                                                            sep='\t',
                                                            index=False,
                                                            header=False)
            if new_val_nodes_for_fold:
                with open(os.path.join(fold_folder_path, ttsConst.NEW_VAL_NODES_FILE_NAME), 'w', newline='\n') as file:
                    file.writelines(list('\n'.join(new_val_nodes_for_fold)))

            i += 1

    def print_vanished_edges(self, vanished_edges):
        self.folder_path = os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME)
        vanished_edges.to_csv(os.path.join(self.folder_path, ttsConst.VANISHED_FILE_NAME),
                              sep='\t',
                              index=False,
                              header=False)



import inspect
import os

from tqdm import tqdm

import openbiolink.graphProperties as graphProp
from openbiolink import globalConfig
from openbiolink import globalConfig as globConst
from openbiolink import utils
from openbiolink.cli import Cli
from openbiolink.graph_creation import graphCreationConfig as gcConst
from openbiolink.graph_creation.file_downloader.fileDownloader import *
from openbiolink.graph_creation.file_processor.fileProcessor import *
from openbiolink.graph_creation.file_reader.fileReader import *
from openbiolink.graph_creation.file_writer.fileWriter import *
from openbiolink.graph_creation.graphCreator import GraphCreator
from openbiolink.graph_creation.graphWriter import GraphWriter
from openbiolink.graph_creation.metadata_db_file import *
from openbiolink.graph_creation.metadata_edge.edgeOntoMetadata import EdgeOntoMetadata
from openbiolink.graph_creation.metadata_edge.edgeRegularMetadata import EdgeRegularMetadata
from openbiolink.graph_creation.metadata_edge.tnEdgeRegularMetadata import TnEdgeRegularMetadata
from openbiolink.graph_creation.metadata_infile import *


class Graph_Creation():
    def __init__(self, folder_path, use_db_metadata_classes=None, use_edge_metadata_classes=None):
        globConst.WORKING_DIR = folder_path
        gcConst.O_FILE_PATH = os.path.join(folder_path, gcConst.O_FILE_FOLDER_NAME)
        gcConst.IN_FILE_PATH = os.path.join(folder_path, gcConst.IN_FILE_FOLDER_NAME)

        if not os.path.exists(globConst.WORKING_DIR):
            os.makedirs(globConst.WORKING_DIR)

        self.db_file_metadata = [x() for x in utils.get_leaf_subclasses(DbMetadata)]
        self.file_readers = [x() for x in utils.get_leaf_subclasses(FileReader)]
        self.file_processors = [x() for x in utils.get_leaf_subclasses(FileProcessor)]
        self.infile_metadata = [x() for x in utils.get_leaf_subclasses(InfileMetadata)]
        self.edge_metadata = [x(graphProp.QUALITY) for x in utils.get_leaf_subclasses(EdgeRegularMetadata)] + \
                             [x(graphProp.QUALITY) for x in utils.get_leaf_subclasses(EdgeOntoMetadata)]
        self.tn_edge_metadata = [x(graphProp.QUALITY) for x in utils.get_leaf_subclasses(TnEdgeRegularMetadata)]

        self.dbType_reader_map = utils.cls_list_to_dic(self.file_readers, 'dbType')
        self.readerType_processor_map = utils.cls_list_to_dic(self.file_processors, 'readerType')
        self.infileType_inMetadata_map = {x.infileType: x for x in self.infile_metadata}

        # if not glob.DIRECTED:
        ## remove onto
        #    if use_edge_metadata_classes is None:
        #        use_edge_metadata_classes = [x(glob.QUALITY) for x in utils.get_leaf_subclasses(EdgeRegularMetadata)]
        #    else:
        #        temp_use_edge_metadata_classes =[]
        #        for edge_class in use_edge_metadata_classes:
        #            if inspect.isclass(edge_class):
        #                if not issubclass(edge_class, EdgeOntoMetadata):
        #                    temp_use_edge_metadata_classes.append(edge_class())
        #            else:
        #                if not issubclass(type(edge_class), EdgeOntoMetadata):
        #                    temp_use_edge_metadata_classes.append(edge_class)
        #        use_edge_metadata_classes = temp_use_edge_metadata_classes
        #        #use_edge_metadata_classes = [x for x in use_edge_metadata_classes if not issubclass(type(x), EdgeOntoMetadata)]

        # use only the desired sources
        if use_db_metadata_classes is not None:
            self.init_custom_sources_bottom_up(use_db_metadata_classes)
        if use_edge_metadata_classes is not None:
            self.init_custom_sources_top_down(use_edge_metadata_classes)

        graphProp.EDGE_TYPES = [str(x.__class__.__name__) for x in self.edge_metadata]
        # testme

    # ----------- download ----------

    def download_db_files(self):
        skip = None
        for_all = False
        if not globalConfig.INTERACTIVE_MODE:
            skip = globalConfig.SKIP_EXISTING_FILES
            for_all = True
        if not os.path.exists(gcConst.O_FILE_PATH):
            os.makedirs(gcConst.O_FILE_PATH)
        for db_file in tqdm(self.db_file_metadata):
            o_file_path = os.path.join(gcConst.O_FILE_PATH, db_file.ofile_name)
            if not for_all:
                if globConst.GUI_MODE:
                    from openbiolink.gui.gui import skipExistingFiles
                    skip, for_all = skipExistingFiles(o_file_path)
                else:
                    skip, for_all = Cli.skip_existing_files(o_file_path)
            if not (skip and os.path.isfile(o_file_path)):
                FileDownloader.download(db_file.url, o_file_path)

    # ----------- create input files ----------

    def create_input_files(self):
        skip = None
        for_all = False
        if not globalConfig.INTERACTIVE_MODE:
            skip = globalConfig.SKIP_EXISTING_FILES
            for_all = True
        if not os.path.exists(gcConst.IN_FILE_PATH):
            os.makedirs(gcConst.IN_FILE_PATH)
        for reader in tqdm(self.file_readers):
            if reader.readerType in self.readerType_processor_map:

                # check beforehand if read in content is processed as parsing can be time consuming
                all_files_exist = True
                for processor in self.readerType_processor_map[reader.readerType]:
                    if not os.path.isfile(os.path.join(gcConst.IN_FILE_PATH, (
                    self.infileType_inMetadata_map[processor.infileType]).csv_name)):
                        all_files_exist = False
                if all_files_exist and not for_all and self.readerType_processor_map[reader.readerType]:
                    first_processor = self.readerType_processor_map[reader.readerType][0]
                    first_processor_out_path = os.path.join(gcConst.IN_FILE_PATH, (
                    self.infileType_inMetadata_map[first_processor.infileType]).csv_name)
                    if globConst.GUI_MODE:
                        from openbiolink.gui.gui import skipExistingFiles
                        skip, for_all = skipExistingFiles(first_processor_out_path)
                    else:
                        skip, for_all = Cli.skip_existing_files(first_processor_out_path)
                if not (skip and all_files_exist):

                    # execute processors
                    in_data = reader.read_file()
                    # fixme  ResourceWarning: Enable tracemalloc to get the object allocation traceback
                    for processor in self.readerType_processor_map[reader.readerType]:
                        out_file_path = os.path.join(gcConst.IN_FILE_PATH,
                                                     (self.infileType_inMetadata_map[processor.infileType]).csv_name)
                        if not for_all:
                            if globConst.GUI_MODE:
                                from openbiolink.gui.gui import skipExistingFiles
                                skip, for_all = skipExistingFiles(out_file_path)
                            else:
                                skip, for_all = Cli.skip_existing_files(out_file_path)
                        if not (skip and os.path.isfile(out_file_path)):
                            out_data = processor.process(in_data)
                            FileWriter.wirte_to_file(out_data, out_file_path)
            else:
                logging.warning('There is no processor for the reader %s' % (str(reader.readerType)))

    # ----------- create graph ----------

    def create_graph(self, one_file_sep='\t', multi_file_sep=None, print_qscore=True):
        gc = GraphCreator()
        gw = GraphWriter()
        # create graph
        nodes_dic, edges_dic = gc.meta_edges_to_graph(self.edge_metadata)
        gw.output_graph(nodes_dic,
                        edges_dic,
                        one_file_sep=one_file_sep,
                        multi_file_sep=multi_file_sep,
                        print_qscore=print_qscore)
        # create TN edges
        tn_nodes_dic, tn_edges_dic = gc.meta_edges_to_graph(self.tn_edge_metadata, tn=True)
        gw.output_graph(tn_nodes_dic,
                        tn_edges_dic,
                        one_file_sep=one_file_sep,
                        multi_file_sep=multi_file_sep,
                        prefix='TN_',
                        print_qscore=print_qscore)
        all_nodes_dic = nodes_dic.copy()
        for key, values in tn_nodes_dic.items():
            if key in all_nodes_dic.keys():
                temp = set(all_nodes_dic[key])
                temp.update(set(values))
                values = temp
            all_nodes_dic[key] = values
        gw.output_graph(all_nodes_dic,
                        None,
                        one_file_sep=one_file_sep,
                        multi_file_sep=None,
                        prefix='ALL_',
                        print_qscore=False,
                        node_edge_list=False
                        )
        graphProp.EDGE_TYPES = list(edges_dic.keys())
        graphProp.NODE_TYPES = list(nodes_dic.keys())

        gw.output_graph_props()

    # ----------- helper init functions ----------

    def init_custom_sources_bottom_up(self, use_db_metdata_classes):
        """helper __init__ function for custom db_metadata_classes"""

        self.db_file_metadata = []

        # remove dbMetadata from list
        # make sure to use instances of classes
        for x in use_db_metdata_classes:
            if inspect.isclass(x):
                self.db_file_metadata.append(x())
            else:
                self.db_file_metadata.append(x)

        # remove readers
        keep_dbType = [x.dbType for x in self.db_file_metadata]
        logging.info(
            'readers removed: ' + str([x.__class__.__name__ for x in self.file_readers if x.dbType not in keep_dbType]))
        self.file_readers = [x for x in self.file_readers if x.dbType in keep_dbType]
        self.dbType_reader_map = utils.cls_list_to_dic(self.file_readers, 'dbType')

        # remove processors
        keep_readerType = [x.readerType for x in self.file_readers]
        logging.info('processors removed: %s' % (
            str([x.__class__.__name__ for x in self.file_processors if x.readerType not in keep_readerType])))
        self.file_processors = [x for x in self.file_processors if x.readerType in keep_readerType]
        self.readerType_processor_map = utils.cls_list_to_dic(self.file_processors, 'readerType')

        # remove infile metadata
        keep_infileType = [x.infileType for x in self.file_processors]
        logging.info('processors removed: ' + str(
            [x.__class__.__name__ for x in self.infile_metadata if x.infileType not in keep_infileType]))
        self.infile_metadata = [x for x in self.infile_metadata if x.infileType in keep_infileType]
        self.infileType_inMetadata_map = {x.infileType: x for x in self.infile_metadata}

        # remove edge metadata
        logging.info('edges removed: ' + str(
            [x.__class__.__name__ for x in self.edge_metadata + self.tn_edge_metadata if
             x.EDGE_INMETA_CLASS.INFILE_TYPE not in keep_infileType]))
        self.edge_metadata = [x for x in self.edge_metadata if x.EDGE_INMETA_CLASS.INFILE_TYPE in keep_infileType]
        self.tn_edge_metadata = [x for x in self.tn_edge_metadata if x.EDGE_INMETA_CLASS.INFILE_TYPE in keep_infileType]

        # check for deleted dependencies of mappings
        additional_remove_metaEdges = []
        additional_remove_mapping_infileType = []
        for metaEdge in self.edge_metadata + self.tn_edge_metadata:
            mappings = [metaEdge.MAP1_META_CLASS, metaEdge.MAP2_META_CLASS, metaEdge.MAP1_ALT_ID_META_CLASS,
                        metaEdge.MAP2_ALT_ID_META_CLASS]
            for mapping in mappings:

                if mapping is not None and mapping.INFILE_TYPE not in keep_infileType:
                    additional_remove_metaEdges.append(metaEdge)
                    additional_remove_mapping_infileType.append(mapping.INFILE_TYPE)
        if len(additional_remove_metaEdges) > 0:
            message = '\nDue to manual exclusion of DB resources, also the edges: %s\n ' \
                      'will be removed due to deleted dependencies of used mappings (i.e. %s\n ' \
                      'Consider manually exclude edges instead of DB resources.' % (
                          str([x.__class__.__name__ for x in additional_remove_metaEdges]),
                          str([str(x) for x in additional_remove_mapping_infileType]))
            logging.warning(message)
            if globConst.GUI_MODE:
                from openbiolink.gui import gui
                gui.askForExit(message)
            elif globConst.INTERACTIVE_MODE:
                Cli.ask_for_exit(message)
            else:
                sys.exit()

            self.edge_metadata = [x for x in self.edge_metadata if x not in additional_remove_metaEdges]
            self.tn_edge_metadata = [x for x in self.tn_edge_metadata if x not in additional_remove_metaEdges]

    def init_custom_sources_top_down(self, use_edge_metdata_classes):
        """hepler __init__ function for custom edge_metadata_classes"""

        # remove edge_metadata
        logging.info('Edge Metadata removed: ' + str([x.__class__.__name__ for x in self.edge_metadata if
                                                      x.EDGE_INMETA_CLASS not in [y.EDGE_INMETA_CLASS for y in
                                                                                  use_edge_metdata_classes]]))
        self.edge_metadata = []

        for x in use_edge_metdata_classes:
            if inspect.isclass(x):
                self.edge_metadata.append(x())
            else:
                self.edge_metadata.append(x)

        # remove inMetadata
        infileType_edgeMetadata_map = utils.cls_list_to_dic(self.edge_metadata, 'EDGE_INMETA_CLASS.INFILE_TYPE')
        infileType_edgeMetadata_map.update(utils.cls_list_to_dic(self.edge_metadata, 'MAP1_META_CLASS.INFILE_TYPE',
                                                                 lambda a: a.MAP1_META_CLASS is not None))
        infileType_edgeMetadata_map.update(utils.cls_list_to_dic(self.edge_metadata, 'MAP2_META_CLASS.INFILE_TYPE',
                                                                 lambda a: a.MAP2_META_CLASS is not None))
        infileType_edgeMetadata_map.update(
            utils.cls_list_to_dic(self.edge_metadata, 'MAP1_ALT_ID_META_CLASS.INFILE_TYPE',
                                  lambda a: a.MAP1_ALT_ID_META_CLASS is not None))
        infileType_edgeMetadata_map.update(
            utils.cls_list_to_dic(self.edge_metadata, 'MAP2_ALT_ID_META_CLASS.INFILE_TYPE',
                                  lambda a: a.MAP2_ALT_ID_META_CLASS is not None))

        infileType_edgeMetadata_map.update(
            utils.cls_list_to_dic(self.tn_edge_metadata, 'EDGE_INMETA_CLASS.INFILE_TYPE'))
        infileType_edgeMetadata_map.update(utils.cls_list_to_dic(self.tn_edge_metadata, 'MAP1_META_CLASS.INFILE_TYPE',
                                                                 lambda a: a.MAP1_META_CLASS is not None))
        infileType_edgeMetadata_map.update(utils.cls_list_to_dic(self.tn_edge_metadata, 'MAP2_META_CLASS.INFILE_TYPE',
                                                                 lambda a: a.MAP2_META_CLASS is not None))
        infileType_edgeMetadata_map.update(
            utils.cls_list_to_dic(self.tn_edge_metadata, 'MAP1_ALT_ID_META_CLASS.INFILE_TYPE',
                                  lambda a: a.MAP1_ALT_ID_META_CLASS is not None))
        infileType_edgeMetadata_map.update(
            utils.cls_list_to_dic(self.tn_edge_metadata, 'MAP2_ALT_ID_META_CLASS.INFILE_TYPE',
                                  lambda a: a.MAP2_ALT_ID_META_CLASS is not None))

        keep_infileTypes = list(infileType_edgeMetadata_map.keys())
        logging.info('Infile Metadata removed: ' + str(
            [x.__class__.__name__ for x in self.infile_metadata if x.infileType not in keep_infileTypes]))
        self.infile_metadata = [x for x in self.infile_metadata if x.infileType in keep_infileTypes]
        self.infileType_inMetadata_map = {x.infileType: x for x in self.infile_metadata}

        # remove processors
        logging.info('Processors removed: ' + str(
            [x.__class__.__name__ for x in self.file_processors if x.infileType not in keep_infileTypes]))
        self.file_processors = [x for x in self.file_processors if x.infileType in keep_infileTypes]
        self.readerType_processor_map = utils.cls_list_to_dic(self.file_processors, 'readerType')

        # remove readers
        keep_readerType = list(self.readerType_processor_map.keys())
        logging.info('Readers removed: ' + str(
            [x.__class__.__name__ for x in self.file_readers if x.readerType not in keep_readerType]))
        self.file_readers = [x for x in self.file_readers if x.readerType in keep_readerType]
        self.dbType_reader_map = utils.cls_list_to_dic(self.file_readers, 'dbType')

        # remove db_metadata
        keep_dbType = list(self.dbType_reader_map.keys())
        logging.info('DB_source removed: ' + str(
            [x.__class__.__name__ for x in self.db_file_metadata if x.dbType not in keep_dbType]))

        self.db_file_metadata = [x for x in self.db_file_metadata if x.dbType in keep_dbType]



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapDisGeNet import InMetaMapDisGeNet
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class MapDisGeNetProcessor(FileProcessor):
    IN_META_CLASS = InMetaMapDisGeNet

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_MAP_DISGENET,
                         infileType=InfileType.IN_MAP_DISGENET, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_preprocessing(self, data):
        # making ids unique in DisGeNet mapping file for DO and OMIM (metadata_db_file:id)
        data.loc[data['voc'] == 'DO', 'code'] = 'DOID:' + data[data['voc'] == 'DO']['code']
        data = data[data['voc'] == 'DO']

        return data



import tkinter as tk
from tkinter import messagebox, ttk

from openbiolink import utils
from openbiolink.graph_creation.metadata_db_file import DbMetadata
from openbiolink.graph_creation.metadata_edge.edgeOntoMetadata import EdgeOntoMetadata
from openbiolink.graph_creation.metadata_edge.edgeRegularMetadata import EdgeRegularMetadata
from openbiolink.gui import gui as gui


class GraphCreationFrame(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller
        self.info_text = ''
        self.db_cls_list = [x for x in utils.get_leaf_subclasses(DbMetadata)]
        self.edge_metadata_cls_list = [x for x in utils.get_leaf_subclasses(EdgeRegularMetadata)] + \
                                      [x for x in utils.get_leaf_subclasses(EdgeOntoMetadata)]
        self.db_cls_list.sort(key=lambda x: x.NAME)
        self.edge_metadata_cls_list.sort(key=lambda x: x.NAME)
        self.selected_dbs = None
        self.selected_meta_edges = None

        titles_panel = tk.Frame(self)
        self.info = tk.Button(titles_panel, text=" help ", command=lambda: gui.show_info_box(self.info_text))
        self.title = tk.Label(titles_panel, text="(1) Graph Creation", font=controller.title_font)

        self.actions_el = self._create_action_el(self)
        select_panel = tk.Frame(self)
        self.select_el = self._create_select_db_meta_edges_el(select_panel)

        options_panel = tk.Frame(self)
        self.graph_prop_el = self._create_graph_prop_el(options_panel)
        self.output_format_el = self._create_output_format_el(options_panel)

        buttons_panel = tk.Frame(self)
        next_button = tk.Button(buttons_panel, text="Next", command=lambda: self.next_page(), height=1, width=15)
        prev_button = tk.Button(buttons_panel, text="Back", command=lambda: self.controller.show_previous_frame(),
                                height=1, width=15)

        titles_panel.pack(side="top", fill="x", pady=10)
        self.title.pack(side='left', pady=10, padx=15)
        self.info.pack(side="right", fill="x", pady=5, padx=15)
        self.actions_el.pack(side='top', fill='both', padx=15, pady=5, expand=True)
        select_panel.pack(side='top', fill='both', padx=10, pady=5, expand=True)
        self.select_el.pack(side='top', fill='both', padx=5, pady=10, expand=True)
        options_panel.pack(side='top', fill='both', padx=5, pady=5, expand=True)
        self.graph_prop_el.pack(side='left', fill='both', padx=10, expand=True)
        self.output_format_el.pack(side='left', fill='both', padx=10, expand=True)

        ttk.Separator(self, orient='horizontal').pack(side='top', fill='x', pady=(15, 0), padx=10, anchor='s')
        buttons_panel.pack(side='bottom', padx=15, fill='x')
        prev_button.pack(side='left', anchor='w', pady=(5, 10))
        next_button.pack(side='right', anchor='e', pady=(5, 10))

    def _create_action_el(self, parent):
        el = tk.LabelFrame(parent, text="Actions")
        self.download = tk.BooleanVar(value=True)
        dl_box = tk.Checkbutton(el, text='perform download', variable=self.download)
        self.create_infiles = tk.BooleanVar(value=True)
        create_in_box = tk.Checkbutton(el, text='create infiles', variable=self.create_infiles)
        self.create_graph = tk.BooleanVar(value=True)
        create_graph_box = tk.Checkbutton(el, text='create graph', variable=self.create_graph,
                                          command=self.toggl_cg_elements)
        dl_box.pack(side='left', padx=5, anchor='w')
        create_in_box.pack(side='left', padx=5, anchor='w')
        create_graph_box.pack(side='left', padx=5, anchor='w')
        return el

    def _create_select_db_meta_edges_el(self, parent):
        el = tk.LabelFrame(parent, text="Customize Graph")
        self.select_choices = ['Default Graph',
                               'Use only subset of Source Databases',
                               'Use only subset of Meta edges']
        self.select = tk.StringVar(value=self.select_choices[0])
        select_menu = tk.OptionMenu(el, self.select, *self.select_choices, command=self.toggl_select_buttons)
        select_menu.config(width=32)
        self.select_db_button = tk.Button(el, text='select data bases...', command=self.select_dbs_popup)
        self.select_meta_edge_button = tk.Button(el, text='select meta edges...', command=self.select_mes_popup)
        select_menu.pack(side='left', padx=5, anchor='w')
        return el

    def select_dbs_popup(self):
        select_popup = tk.Toplevel()
        select_popup.wm_title("Select Source Databases")
        select_popup.wm_geometry('400x400')
        label = tk.Label(select_popup, text='Data Base Type - Data Base - Data Base File')
        filterPanel = tk.Frame(select_popup)
        select_all_button = tk.Button(filterPanel, text='select all', command=self.select_all_dbs)
        select_none_button = tk.Button(filterPanel, text='select none', command=self.unselect_all_dbs)
        frame = tk.Frame(select_popup)
        scrollbar = tk.Scrollbar(frame)
        self.db_selection = tk.Listbox(frame, selectmode='multiple', exportselection=0, yscrollcommand=scrollbar.set)
        for i, source_db_cls in enumerate(self.db_cls_list):
            self.db_selection.insert(i, source_db_cls.NAME)
            if source_db_cls in self.selected_dbs:
                self.db_selection.selection_set(i)

        scrollbar.config(command=self.db_selection.yview)
        ok_button = ttk.Button(select_popup, text="Okay", command=lambda: self.safe_select_dbs_and_quit(select_popup))

        label.pack(side='top', anchor='w', pady=10, padx=5)
        filterPanel.pack(side='top')
        select_all_button.pack(side='left')
        select_none_button.pack(side='left')
        frame.pack(side='top', fill='both', expand=True)
        scrollbar.pack(side='right', fill='y')
        self.db_selection.pack(fill='both', padx=10, expand=True)
        ok_button.pack(side='bottom')

    def select_all_dbs(self):
        for i, _ in enumerate(self.db_cls_list):
            self.db_selection.selection_set(i)

    def unselect_all_dbs(self):
        self.db_selection.selection_clear(0, len(self.db_cls_list) - 1)

    def safe_select_dbs_and_quit(self, pop_up):
        self.selected_dbs = [self.db_cls_list[i] for i in self.db_selection.curselection()]
        pop_up.destroy()

    def select_mes_popup(self):
        select_popup = tk.Toplevel()
        select_popup.wm_title("Select Meta Edges")
        select_popup.wm_geometry('400x400')
        label = tk.Label(select_popup, text='Connection Type - Meta Edge Type')
        filterPanel = tk.Frame(select_popup)
        select_all_button = tk.Button(filterPanel, text='select all', command=self.select_all_mes)
        select_none_button = tk.Button(filterPanel, text='select none', command=self.unselect_all_mes)
        frame = tk.Frame(select_popup)
        scrollbar = tk.Scrollbar(frame)

        self.me_selection = tk.Listbox(frame, selectmode='multiple', exportselection=0, yscrollcommand=scrollbar.set)
        for i, source_me_cls in enumerate(self.edge_metadata_cls_list):
            self.me_selection.insert(i, source_me_cls.NAME)
            if source_me_cls in self.selected_meta_edges:
                self.me_selection.selection_set(i)
        scrollbar.config(command=self.me_selection.yview)
        ok_button = ttk.Button(select_popup, text="Okay", command=lambda: self.safe_select_mes_and_quit(select_popup))

        label.pack(side='top', anchor='w', pady=10, padx=5)
        filterPanel.pack(side='top')
        select_all_button.pack(side='left')
        select_none_button.pack(side='left')
        frame.pack(side='top', fill='both', expand=True)
        scrollbar.pack(side='right', fill='y')
        self.me_selection.pack(fill='both', padx=10, expand=True)
        ok_button.pack(side='bottom')

    def safe_select_mes_and_quit(self, pop_up):
        self.selected_meta_edges = [self.edge_metadata_cls_list[i] for i in self.me_selection.curselection()]
        pop_up.destroy()

    def select_all_mes(self):
        for i, _ in enumerate(self.edge_metadata_cls_list):
            self.me_selection.selection_set(i)

    def unselect_all_mes(self):
        self.me_selection.selection_clear(0, len(self.edge_metadata_cls_list) - 1)

    def toggl_select_buttons(self, select_choice):
        if select_choice == self.select_choices[0]:
            self.selected_dbs = None
            self.selected_meta_edges = None
            self.select_db_button.pack_forget()
            self.select_meta_edge_button.pack_forget()
        elif select_choice == self.select_choices[1]:
            self.selected_dbs = self.db_cls_list.copy()
            self.selected_meta_edges = None
            self.select_db_button.pack(side='right', padx=5, anchor='w')
            self.select_meta_edge_button.pack_forget()
            self.select_dbs_popup()
        elif select_choice == self.select_choices[2]:
            self.selected_dbs = None
            self.selected_meta_edges = self.edge_metadata_cls_list.copy()
            self.select_db_button.pack_forget()
            self.select_meta_edge_button.pack(side='right', padx=5, anchor='w')
            self.select_mes_popup()

    def _create_graph_prop_el(self, parent):
        el = tk.LabelFrame(parent, text="Graph Properties")
        # quality
        self.qual = tk.StringVar(value='None')
        hq_box = tk.Radiobutton(el, text="high quality", variable=self.qual, value='hq')
        mq_box = tk.Radiobutton(el, text="medium quality", variable=self.qual, value='mq')
        lq_box = tk.Radiobutton(el, text="low quality", variable=self.qual, value='lq')
        none_box = tk.Radiobutton(el, text="no quality cutoff", variable=self.qual, value='None')
        # undirected
        self.undir = tk.BooleanVar()
        undirected_box = tk.Radiobutton(el, text='undirected', variable=self.undir, value=True)
        directed_box = tk.Radiobutton(el, text='directed', variable=self.undir, value=False)
        # packing
        none_box.pack(side='top', padx=5, anchor='w')
        hq_box.pack(side='top', padx=5, anchor='w')
        mq_box.pack(side='top', padx=5, anchor='w')
        lq_box.pack(side='top', padx=5, anchor='w')
        ttk.Separator(el, orient='horizontal').pack(side='top', fill='x', pady=2, padx=5, anchor='w')
        directed_box.pack(side='top', pady=2, padx=5, anchor='w')
        undirected_box.pack(side='top', pady=2, padx=5, anchor='w')
        return el

    def _create_output_format_el(self, parent):
        el = tk.LabelFrame(parent, text="Output Format")
        # single outputfile
        self.one_output_file = tk.BooleanVar(value=True)
        single_out_file_box = tk.Checkbutton(el, text='single file', variable=self.one_output_file)
        self.single_sep = tk.StringVar(value='t')
        single_sep_frame = tk.Frame(el)
        single_sep_label = tk.Label(single_sep_frame, text='separator:')
        single_sep_info = tk.Label(single_sep_frame, text='(t for tab, n for newline)', font=self.controller.info_font)
        single_sep_value = tk.Entry(single_sep_frame, textvariable=self.single_sep, width=5)
        # multiple output files
        self.multi_output_file = tk.BooleanVar(value=False)
        multi_out_file_box = tk.Checkbutton(el, text='multiple files (one/type)', variable=self.multi_output_file)
        multi_sep_frame = tk.Frame(el)
        self.multi_sep = tk.StringVar(value=None)
        multi_sep_label = tk.Label(multi_sep_frame, text='separator:')
        multi_sep_info = tk.Label(multi_sep_frame, text='(t for tab, n for newline)', font=self.controller.info_font)
        multi_sep_value = tk.Entry(multi_sep_frame, textvariable=self.multi_sep, width=5)
        # qscore
        self.no_qscore = tk.BooleanVar(value=False)
        no_qscore_box = tk.Checkbutton(el, text='without quality score', variable=self.no_qscore)
        # packing
        single_out_file_box.pack(side='top', padx=5, anchor='w')
        single_sep_frame.pack(side='top', padx=5, anchor='w')
        single_sep_label.pack(side='left', padx=5, anchor='w')
        single_sep_value.pack(side='left', anchor='w')
        single_sep_info.pack(side='left', anchor='w')
        ttk.Separator(el, orient='horizontal').pack(side='top', fill='x', pady=5, padx=5, anchor='w')
        multi_out_file_box.pack(side='top', padx=5, anchor='w')
        multi_sep_frame.pack(side='top', padx=5, anchor='w')
        multi_sep_label.pack(side='left', padx=5, anchor='w')
        multi_sep_value.pack(side='left', anchor='w')
        multi_sep_info.pack(side='left', anchor='w')
        ttk.Separator(el, orient='horizontal').pack(side='top', fill='x', pady=5, padx=5, anchor='w')
        no_qscore_box.pack(side='top', padx=5, anchor='w')
        return el

    def toggl_cg_elements(self):
        if self.graph_prop_el.winfo_ismapped():
            self.select_el.pack_forget()
            self.graph_prop_el.pack_forget()
            self.output_format_el.pack_forget()
        else:
            self.select_el.pack(side='top', fill='both', padx=5, expand=True)
            self.graph_prop_el.pack(side='left', fill='both', padx=10, expand=True)
            self.output_format_el.pack(side='left', fill='both', padx=10, expand=True)

    def next_page(self):
        self.controller.ARGS_LIST_GRAPH_CREATION = []
        if (self.one_output_file.get() and self.single_sep.get() == '') or (
                self.multi_output_file.get() and self.multi_sep.get() == ''):
            messagebox.showerror('ERROR', 'Please provide a separator for desired output file')
            return
        self.controller.ARGS_LIST_GRAPH_CREATION.append('-g')
        if self.undir.get():
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--undir')
        if not self.qual.get() == 'None':
            self.controller.ARGS_LIST_GRAPH_CREATION.extend(['--qual', self.qual.get()])
        if not self.download.get():
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--no_dl')
        if not self.create_infiles.get():
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--no_in')
        if not self.create_graph.get():
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--no_create')
        if self.one_output_file.get():
            if self.multi_output_file.get():
                self.controller.ARGS_LIST_GRAPH_CREATION.extend(
                    ['--out_format', 'sm', self.single_sep.get() + self.multi_sep.get()])
            else:
                self.controller.ARGS_LIST_GRAPH_CREATION.extend(['--out_format', 's', self.single_sep.get()])
        elif self.multi_output_file.get():
            self.controller.ARGS_LIST_GRAPH_CREATION.extend(['--out_format', 'm', self.multi_sep.get()])
        if self.no_qscore.get():
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--no_qscore')
        if self.selected_dbs:
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--dbs')
            self.controller.ARGS_LIST_GRAPH_CREATION.extend(
                [x.__module__ + '.' + x.__name__ for x in self.selected_dbs])
        if self.selected_meta_edges:
            self.controller.ARGS_LIST_GRAPH_CREATION.append('--mes')
            self.controller.ARGS_LIST_GRAPH_CREATION.extend(
                [x.__module__ + '.' + x.__name__ for x in self.selected_meta_edges])

        self.controller.show_next_frame()



from openbiolink.edgeType import EdgeType
from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.nodeType import NodeType


class InMetaOntoGoPartOf(InfileMetadata):
    CSV_NAME = "DB_ONTO_GO_PART_OF_ontology.csv"
    USE_COLS = ['ID', 'PART_OF']
    NODE1_COL = 0
    NODE2_COL = 1
    QSCORE_COL = None
    NODE1_TYPE = NodeType.GO
    NODE2_TYPE = NodeType.GO
    EDGE_TYPE = EdgeType.PART_OF
    MAPPING_SEP = ';'
    INFILE_TYPE = InfileType.IN_ONTO_GO_PART_OF

    def __init__(self):
        super().__init__(csv_name=self.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=self.INFILE_TYPE)



import multiprocessing
import os
import random
from multiprocessing import Pool

import numpy as np
import pandas
from sortedcontainers import SortedList
from tqdm import tqdm

import openbiolink.evaluation.evalConfig as evalConst
import openbiolink.evaluation.evaluationIO as io
from openbiolink import globalConfig as globConst
from openbiolink import utils
from openbiolink.evaluation.metricTypes import RankMetricType, ThresholdMetricType
from openbiolink.evaluation.models.model import Model


class Evaluation:
    def __init__(self, model: Model, training_set_path=None, test_set_path=None, nodes_path=None, mappings_avail=False):
        self.model = model
        if training_set_path:
            self.training_examples = pandas.read_csv(training_set_path, sep="\t", names=globConst.COL_NAMES_SAMPLES)
        else:
            self.training_examples = pandas.DataFrame(columns=globConst.COL_NAMES_SAMPLES)
        if test_set_path:
            self.test_examples = pandas.read_csv(test_set_path, sep="\t", names=globConst.COL_NAMES_SAMPLES)
        else:
            self.test_examples = pandas.DataFrame(columns=globConst.COL_NAMES_SAMPLES)
        if nodes_path is not None:
            self.nodes = pandas.read_csv(nodes_path, sep="\t", names=globConst.COL_NAMES_NODES)
        else:
            self.nodes = None

        if not mappings_avail:
            relation_labels = set()
            relation_labels.update(set(self.test_examples[globConst.EDGE_TYPE_COL_NAME]))
            relation_labels.update(set(self.training_examples[globConst.EDGE_TYPE_COL_NAME]))

            self.node_label_to_id = None
            self.node_types_to_id = None
            self.relation_label_to_id = utils.create_mappings(relation_labels)
            if self.nodes is not None:
                self.node_label_to_id = utils.create_mappings(np.unique(self.nodes.values[:, 0]))
                self.node_types_to_id = utils.create_mappings(np.unique(self.nodes.values[:, 1]))
            else:
                node_labels = set()
                node_labels.update(set(self.test_examples[globConst.NODE1_ID_COL_NAME]))
                node_labels.update(set(self.test_examples[globConst.NODE2_ID_COL_NAME]))
                node_labels.update(set(self.training_examples[globConst.NODE1_ID_COL_NAME]))
                node_labels.update(set(self.training_examples[globConst.NODE2_ID_COL_NAME]))
                self.node_label_to_id = utils.create_mappings(node_labels)

            # output mappings
            io.write_mappings(node_label_to_id=self.node_label_to_id,
                              node_types_to_id=self.node_types_to_id,
                              relation_label_to_id=self.relation_label_to_id)

        else:
            # testme
            output_directory = os.path.join(os.path.join(globConst.WORKING_DIR, evalConst.EVAL_OUTPUT_FOLDER_NAME),
                                            evalConst.MODEL_DIR)
            self.node_label_to_id = io.read_mapping(
                os.path.join(output_directory, evalConst.MODEL_ENTITY_NAME_MAPPING_NAME))
            node_type_path = os.path.join(output_directory, evalConst.MODEL_ENTITY_TYPE_MAPPING_NAME)
            if os.path.exists(node_type_path):
                self.node_types_to_id = io.read_mapping(node_type_path)
            elif self.nodes is not None:
                self.node_types_to_id = utils.create_mappings(np.unique(self.nodes.values[:, 1]))
                io.write_mappings(node_types_to_id=self.node_types_to_id)
            else:
                pass  # fixme error nodes must be provided when eval
            self.relation_label_to_id = io.read_mapping(
                os.path.join(output_directory, evalConst.MODEL_RELATION_TYPE_MAPPING_NAME))

    def train(self):
        ### prepare input examples
        pos_examples = self.training_examples[self.training_examples[globConst.VALUE_COL_NAME] == 1]
        neg_examples = self.training_examples[self.training_examples[globConst.VALUE_COL_NAME] == 0]
        # pos and neg must be same length! but also, all entities must still be present!
        num_examples = min(len(neg_examples), len(pos_examples))
        pos_examples = self.save_remove_n_edges(pos_examples, len(pos_examples) - num_examples)
        neg_examples = self.save_remove_n_edges(neg_examples, len(neg_examples) - num_examples)
        pos_triples = pos_examples[globConst.COL_NAMES_TRIPLES].values
        neg_triples = neg_examples[globConst.COL_NAMES_TRIPLES].values
        mapped_pos_triples, _ = self.get_mapped_triples_and_nodes(triples=pos_triples)
        mapped_neg_triples, _ = self.get_mapped_triples_and_nodes(triples=neg_triples)

        self.model.train(pos_triples=mapped_pos_triples, neg_triples=mapped_neg_triples)
        output_directory = os.path.join(os.path.join(globConst.WORKING_DIR, evalConst.EVAL_OUTPUT_FOLDER_NAME),
                                        evalConst.MODEL_DIR)

        os.makedirs(output_directory, exist_ok=True)
        model_path = os.path.join(output_directory, evalConst.MODEL_TRAINED_NAME)
        self.model.output_model(model_path)

    def evaluate(self, metrics: list, ks=None):
        if not ks:
            ks = evalConst.DEFAULT_HITS_AT_K
        os.makedirs(os.path.join(globConst.WORKING_DIR, evalConst.EVAL_OUTPUT_FOLDER_NAME), exist_ok=True)

        threshold_metrics = [m for m in ThresholdMetricType]
        num_threshold_metrics = len([x for x in threshold_metrics if x in metrics])
        ranked_metrics = [m for m in RankMetricType]
        num_ranked_metrics = len([x for x in ranked_metrics if x in metrics])

        unfiltered_metrics = [RankMetricType.MRR_UNFILTERED, RankMetricType.HITS_AT_K_UNFILTERED]
        num_ranked_unfiltered_metrics = len([x for x in unfiltered_metrics if x in metrics])
        filtered_options = bool(num_ranked_metrics - num_ranked_unfiltered_metrics)
        unfiltered_options = bool(num_ranked_unfiltered_metrics)
        metrics_results = {}

        if num_ranked_metrics > 0:
            ranked_metrics_results = self.evaluate_ranked_metrics_2(metrics=metrics,
                                                                    ks=ks,
                                                                    filtered_setting=filtered_options,
                                                                    unfiltered_setting=unfiltered_options)
            metrics_results.update(ranked_metrics_results)

        if num_threshold_metrics > 0:
            threshold_metrics_results = self.evaluate_threshold_metrics(metrics=metrics)
            metrics_results.update(threshold_metrics_results)

        io.write_metric_results(metrics_results)

        return metrics_results

    # ---------------------------- START -----------------------------------------
    def evaluate_ranked_metrics_3(self, ks, metrics, unfiltered_setting=True, filtered_setting=False):
        metric_results = {}
        k_raw_corrupted_head = []
        for _ in ks:
            k_raw_corrupted_head.append([])
        k_raw_corrupted_tail = []
        for _ in ks:
            k_raw_corrupted_tail.append([])

        # get corrupted triples
        pos_test_examples = self.test_examples[self.test_examples[globConst.VALUE_COL_NAME] == 1]
        pos_test_examples_array = pos_test_examples.values
        nodes_array = self.nodes.values

        mapped_pos_triples, mapped_nodes = self.get_mapped_triples_and_nodes(triples=pos_test_examples_array,
                                                                             nodes=nodes_array)
        nodeTypes = np.unique(mapped_nodes[:, 1])
        nodes_dic = {nodeType: np.unique(mapped_nodes[np.where(mapped_nodes[:, 1] == nodeType)][:, 0]) for nodeType in
                     nodeTypes}
        head_tuples = mapped_pos_triples[:, 0:2]
        head_tuples = np.unique(head_tuples, axis=0)
        tail_tuples = mapped_pos_triples[:, 1:3]
        tail_tuples = np.unique(tail_tuples, axis=0)
        head_ranks = []
        # corrupting tail
        for head, relation in tqdm(head_tuples):
            data = mapped_pos_triples[np.where((mapped_pos_triples[:, 0] == head) *
                                               (mapped_pos_triples[:, 1] == relation))]

            ranked_pos_examples, _ = self.model.get_ranked_and_sorted_predictions(data)
            _, corrupted_examples, _, _ = utils.calc_corrupted_triples(
                pos_example=data[0],
                nodes=mapped_nodes,
                nodes_dic=nodes_dic,
                filtered=False,
                pos_examples=mapped_pos_triples)
            all_examples = np.unique(np.row_stack((corrupted_examples, np.column_stack((data, [0] * len(data))))),
                                     axis=0)  # todo VERY WRONG!
            ranked_all_examples, _ = self.model.get_ranked_and_sorted_predictions(all_examples)
            increase_search_frame_by = [0] * len(ks)
            for example in ranked_pos_examples:
                search_data = ranked_all_examples[0:ks[-1] + 1, :]  # fixme this should be more?
                for i, k in enumerate(ks):
                    current_k = k + increase_search_frame_by[i]
                    current_k = min(current_k, len(search_data))
                    index = np.where(search_data[:, 2] == example[2])[0]
                    if index <= current_k:
                        k_raw_corrupted_tail[i].append(1)
                        increase_search_frame_by[i] += 1
                    else:
                        k_raw_corrupted_tail[i].append(0)

        # corrupting head
        for relation, tail in tqdm(tail_tuples):
            data = mapped_pos_triples[np.where((mapped_pos_triples[:, 1] == relation) *
                                               (mapped_pos_triples[:, 2] == tail))]

            ranked_pos_examples, _ = self.model.get_ranked_and_sorted_predictions(data)
            corrupted_examples, _, _, _ = utils.calc_corrupted_triples(
                pos_example=data[0],
                nodes=mapped_nodes,
                nodes_dic=nodes_dic,
                filtered=False,
                pos_examples=mapped_pos_triples)
            all_examples = np.unique(np.row_stack((corrupted_examples, np.column_stack((data, [0] * len(data))))),
                                     axis=0)  # todo VERY WRONG!
            ranked_all_examples, _ = self.model.get_ranked_and_sorted_predictions(all_examples)
            increase_search_frame_by = [0] * len(ks)
            for example in ranked_pos_examples:
                search_data = ranked_all_examples[0:ks[-1] + 1, :]
                for i, k in enumerate(ks):
                    current_k = k + increase_search_frame_by[i]
                    current_k = min(current_k, len(search_data))
                    index = np.where(search_data[:, 0] == example[0])[0] + 1
                    if index <= current_k:
                        k_raw_corrupted_head[i].append(1)
                        increase_search_frame_by[i] += 1
                    else:
                        k_raw_corrupted_head[i].append(0)
        k_results_corrupted_head = []
        for i, k in enumerate(ks):
            k_results_corrupted_head.append(sum(k_raw_corrupted_head[i]) / len(k_raw_corrupted_head[i]))
        k_results_corrupted_tail = []
        for i, k in enumerate(ks):
            k_results_corrupted_tail.append(sum(k_raw_corrupted_tail[i]) / len(k_raw_corrupted_tail[i]))

        metric_results[RankMetricType.HITS_AT_K] = (k_results_corrupted_head, k_results_corrupted_tail)
        return metric_results

    def evaluate_ranked_metrics_2(self, ks, metrics, unfiltered_setting=True, filtered_setting=False):
        metric_results = {}

        # get corrupted triples
        pos_test_examples = self.test_examples[self.test_examples[globConst.VALUE_COL_NAME] == 1]
        pos_test_examples_array = pos_test_examples.values
        nodes_array = self.nodes.values

        mapped_pos_triples, mapped_nodes = self.get_mapped_triples_and_nodes(triples=pos_test_examples_array,
                                                                             nodes=nodes_array)

        node_types = np.unique(mapped_nodes[:, 1])
        nodes_dic = {nodeType: mapped_nodes[np.where(mapped_nodes[:, 1] == nodeType)][:, 0] for nodeType in node_types}

        print('calculating corrupted triples')

        p = Pool(processes=multiprocessing.cpu_count() - 1)
        params = [(pos_example, mapped_nodes, nodes_dic, mapped_pos_triples, filtered_setting, unfiltered_setting) for
                  pos_example in mapped_pos_triples]
        # print(params) #todo here
        rank_lists = p.map(self.get_rank_lists, params)
        filtered_ranks_corrupted_heads = [filtered_head for
                                          (unfiltered_head, unfiltered_tail, filtered_head, filtered_tail) in
                                          rank_lists]
        filtered_ranks_corrupted_tails = [filtered_tail for
                                          (unfiltered_head, unfiltered_tail, filtered_head, filtered_tail) in
                                          rank_lists]
        unfiltered_ranks_corrupted_heads = [unfiltered_head for
                                            (unfiltered_head, unfiltered_tail, filtered_head, filtered_tail) in
                                            rank_lists]
        unfiltered_ranks_corrupted_tails = [unfiltered_tail for
                                            (unfiltered_head, unfiltered_tail, filtered_head, filtered_tail) in
                                            rank_lists]

        filtered_num_examples = len(filtered_ranks_corrupted_heads)
        unfiltered_num_examples = len(unfiltered_ranks_corrupted_heads)

        # HITS@K
        if RankMetricType.HITS_AT_K in metrics:
            metric_results[RankMetricType.HITS_AT_K] = self.calculate_hits_at_k(ks=ks,
                                                                                ranks_corrupted_heads=filtered_ranks_corrupted_heads,
                                                                                ranks_corrupted_tails=filtered_ranks_corrupted_tails,
                                                                                num_examples=filtered_num_examples)
        # HITS@K unfiltered
        if RankMetricType.HITS_AT_K_UNFILTERED in metrics:
            metric_results[RankMetricType.HITS_AT_K_UNFILTERED] = self.calculate_hits_at_k(ks=ks,
                                                                                           ranks_corrupted_heads=unfiltered_ranks_corrupted_heads,
                                                                                           ranks_corrupted_tails=unfiltered_ranks_corrupted_tails,
                                                                                           num_examples=unfiltered_num_examples)
        # MRR
        if RankMetricType.MRR in metrics:
            metric_results[RankMetricType.MRR] = self.calculate_mrr(
                ranks_corrupted_heads=filtered_ranks_corrupted_heads,
                ranks_corrupted_tails=filtered_ranks_corrupted_tails,
                num_examples=filtered_num_examples)
        # MRR unfiltered
        if RankMetricType.MRR_UNFILTERED in metrics:
            metric_results[RankMetricType.MRR] = self.calculate_mrr(
                ranks_corrupted_heads=unfiltered_ranks_corrupted_heads,
                ranks_corrupted_tails=unfiltered_ranks_corrupted_tails,
                num_examples=unfiltered_num_examples)
        return metric_results

    def evaluate_ranked_metrics_1(self, ks, metrics, unfiltered_setting=True, filtered_setting=False):
        metric_results = {}

        # get corrupted triples
        pos_test_examples = self.test_examples[self.test_examples[globConst.VALUE_COL_NAME] == 1]
        pos_test_examples_array = pos_test_examples.values
        nodes_array = self.nodes.values

        mapped_pos_triples, mapped_nodes = self.get_mapped_triples_and_nodes(triples=pos_test_examples_array,
                                                                             nodes=nodes_array)

        node_types = np.unique(mapped_nodes[:, 1])
        nodes_dic = {nodeType: mapped_nodes[np.where(mapped_nodes[:, 1] == nodeType)][:, 0] for nodeType in node_types}

        filtered_ranks_corrupted_heads = []
        filtered_ranks_corrupted_tails = []
        unfiltered_ranks_corrupted_heads = []
        unfiltered_ranks_corrupted_tails = []

        print('calculating corrupted triples')

        for pos_example in tqdm(mapped_pos_triples, total=mapped_pos_triples.shape[0]):
            unfiltered_corrupted_head, \
            unfiltered_corrupted_tail, \
            filtered_corrupted_head, \
            filtered_corrupted_tail = utils.calc_corrupted_triples(
                pos_example=pos_example,
                nodes=mapped_nodes,
                nodes_dic=nodes_dic,
                filtered=filtered_setting,
                pos_examples=mapped_pos_triples)
            if unfiltered_setting:
                unfiltered_ranks_corrupted_heads.append(
                    self.get_rank_for_corrupted_examples(unfiltered_corrupted_head, pos_example))
                unfiltered_ranks_corrupted_tails.append(
                    self.get_rank_for_corrupted_examples(unfiltered_corrupted_tail, pos_example))
            if filtered_setting:
                filtered_ranks_corrupted_heads.append(
                    self.get_rank_for_corrupted_examples(filtered_corrupted_head, pos_example))
                filtered_ranks_corrupted_tails.append(
                    self.get_rank_for_corrupted_examples(filtered_corrupted_tail, pos_example))

        filtered_num_examples = len(filtered_ranks_corrupted_heads)
        unfiltered_num_examples = len(unfiltered_ranks_corrupted_heads)

        # HITS@K
        if RankMetricType.HITS_AT_K in metrics:
            metric_results[RankMetricType.HITS_AT_K] = self.calculate_hits_at_k(ks=ks,
                                                                                ranks_corrupted_heads=filtered_ranks_corrupted_heads,
                                                                                ranks_corrupted_tails=filtered_ranks_corrupted_tails,
                                                                                num_examples=filtered_num_examples)
        # HITS@K unfiltered
        if RankMetricType.HITS_AT_K_UNFILTERED in metrics:
            metric_results[RankMetricType.HITS_AT_K_UNFILTERED] = self.calculate_hits_at_k(ks=ks,
                                                                                           ranks_corrupted_heads=unfiltered_ranks_corrupted_heads,
                                                                                           ranks_corrupted_tails=unfiltered_ranks_corrupted_tails,
                                                                                           num_examples=unfiltered_num_examples)
        # MRR
        if RankMetricType.MRR in metrics:
            metric_results[RankMetricType.MRR] = self.calculate_mrr(
                ranks_corrupted_heads=filtered_ranks_corrupted_heads,
                ranks_corrupted_tails=filtered_ranks_corrupted_tails,
                num_examples=filtered_num_examples)
        # MRR unfiltered
        if RankMetricType.MRR_UNFILTERED in metrics:
            metric_results[RankMetricType.MRR] = self.calculate_mrr(
                ranks_corrupted_heads=unfiltered_ranks_corrupted_heads,
                ranks_corrupted_tails=unfiltered_ranks_corrupted_tails,
                num_examples=unfiltered_num_examples)
        return metric_results

    def get_rank_lists(self, params):
        unfiltered_head_ranks = None
        unfiltered_tail_ranks = None
        filtered_head_ranks = None
        filtered_tail_ranks = None
        pos_example, mapped_nodes, nodes_dic, mapped_pos_triples, filtered_setting, unfiltered_setting = params
        unfiltered_corrupted_head, \
        unfiltered_corrupted_tail, \
        filtered_corrupted_head, \
        filtered_corrupted_tail = utils.calc_corrupted_triples(
            pos_example=pos_example,
            nodes=mapped_nodes,
            nodes_dic=nodes_dic,
            filtered=filtered_setting,
            pos_examples=mapped_pos_triples)
        if unfiltered_setting:
            unfiltered_head_ranks = self.get_rank_for_corrupted_examples(unfiltered_corrupted_head, pos_example)
            unfiltered_tail_ranks = self.get_rank_for_corrupted_examples(unfiltered_corrupted_tail, pos_example)
        if filtered_setting:
            filtered_head_ranks = self.get_rank_for_corrupted_examples(filtered_corrupted_head, pos_example)
            filtered_tail_ranks = self.get_rank_for_corrupted_examples(filtered_corrupted_tail, pos_example)
        return unfiltered_head_ranks, unfiltered_tail_ranks, filtered_head_ranks, filtered_tail_ranks

    def evaluate_threshold_metrics(self, metrics):
        metric_results = {}
        mapped_test_examples, _ = self.get_mapped_triples_and_nodes(
            triples=self.test_examples.values)  # , nodes=nodes_array) #todo change here
        values = self.test_examples.values[:, 4].tolist()
        mapped_test_examples = np.column_stack((mapped_test_examples, values))
        ranked_test_examples, sorted_indices = self.model.get_ranked_and_sorted_predictions(mapped_test_examples)
        ranked_scores = ranked_test_examples[:, 4].tolist()
        ranked_labels = ranked_test_examples[:, 3].tolist()  # todo change here!!
        # ROC Curve
        if ThresholdMetricType.ROC in metrics:
            fpr, tpr = self.calculate_roc_curve(labels=ranked_labels, scores=ranked_scores)
            metric_results[ThresholdMetricType.ROC] = (fpr, tpr)
        # Precision Recall Curve
        if ThresholdMetricType.PR_REC_CURVE in metrics:
            pr, rec = self.calculate_pr_curve(ranked_labels, ranked_scores)
            metric_results[ThresholdMetricType.PR_REC_CURVE] = (pr, rec)
        # ROC AUC
        if ThresholdMetricType.ROC_AUC:
            if ThresholdMetricType.ROC in metric_results.keys():
                fpr, tpr = metric_results[ThresholdMetricType.ROC]
            else:
                fpr, tpr = self.calculate_roc_curve(labels=ranked_labels, scores=ranked_scores)
                # todo ? auch unique?
            roc_auc = self.calculate_auc(fpr, tpr)
            metric_results[ThresholdMetricType.ROC_AUC] = roc_auc
        # Precision Recall AUC
        if ThresholdMetricType.PR_AUC:
            if ThresholdMetricType.PR_AUC in metric_results.keys():
                pr, rec = metric_results[ThresholdMetricType.PR_REC_CURVE]
            else:
                pr, rec = self.calculate_pr_curve(labels=ranked_labels, scores=ranked_scores)
                pr = np.asarray(pr)
                rec = np.asarray(rec)
            _, indices = np.unique(pr, return_index=True)
            pr_unique = pr[indices]
            rec_unique = rec[indices]
            pr_auc = self.calculate_auc(pr_unique, rec_unique)
            metric_results[ThresholdMetricType.PR_AUC] = pr_auc
        return metric_results

    def get_rank_for_corrupted_examples(self, corrupted_examples, true_triple):
        corrupted_examples = np.row_stack((corrupted_examples, (list(true_triple) + [1])))
        ranked_examples, sorted_indices = self.model.get_ranked_and_sorted_predictions(corrupted_examples)
        # testme
        # ranked_examples.reset_index(drop=True, inplace=True)
        # ranked_labels = corrupted_examples[globConst.VALUE_COL_NAME][sorted_indices]
        return self.get_first_positive_rank(ranked_examples[:, 3])

    def get_mapped_triples_and_nodes(self, triples=None, nodes=None):
        mapped_triples = None
        mapped_nodes = None
        if nodes is not None:
            mapped_nodes = np.column_stack((utils.map_elements(nodes[:, 0:1], self.node_label_to_id),
                                            utils.map_elements(nodes[:, 1:2], self.node_types_to_id)))
        if triples is not None:
            mapped_triples = np.column_stack((utils.map_elements(triples[:, 0:1], self.node_label_to_id),
                                              utils.map_elements(triples[:, 1:2], self.relation_label_to_id),
                                              utils.map_elements(triples[:, 2:3], self.node_label_to_id)))
        return mapped_triples, mapped_nodes

    def create_mappings(self, relations, node_types, node_labels=None):
        self.node_label_to_id = utils.create_mappings(node_labels)
        self.node_types_to_id = utils.create_mappings(node_types)
        self.relation_label_to_id = utils.create_mappings(relations)

    @staticmethod
    def get_first_positive_rank(labels):
        rank = next(i for i, l in enumerate(labels) if l == 1)
        return rank + 1

    @staticmethod
    def save_remove_n_edges(edges: pandas.DataFrame, n):
        """
        removes n edges from 'edges' so that no node is removed in the process, i.e. the total number
        of nodes in 'edges' stays the same
        :param edges: original edges
        :param n: number of how many edges should be removed
        :return:
        """
        if n < 1:
            return edges
        all_edges = SortedList(list(edges[globConst.NODE1_ID_COL_NAME].append(edges[globConst.NODE2_ID_COL_NAME])))
        edges_list = set(all_edges)
        edges_count_dict = {x: all_edges.count(x) for x in edges_list}
        i = 0

        for _ in range(1000):
            drop_indices_candidates = random.sample(edges.index.values.tolist(), n)
            drop_indices = []
            for drop_index in tqdm(drop_indices_candidates):
                if i == n:
                    break
                drop_edge_candidate = edges.loc[drop_index]
                if edges_count_dict[drop_edge_candidate[globConst.NODE1_ID_COL_NAME]] > 1 \
                        and edges_count_dict[drop_edge_candidate[globConst.NODE2_ID_COL_NAME]] > 1:
                    drop_indices.append(drop_index)
                    i += 1
            edges.drop(inplace=True, index=drop_indices)
            if i == n:
                break
        edges.reset_index(drop=True, inplace=True)
        return edges

    ###### calculate metrics ######

    @staticmethod
    def calculate_hits_at_k(ks, ranks_corrupted_heads, ranks_corrupted_tails, num_examples):
        corrupted_heads_hits_at_k = dict()
        corrupted_tails_hits_at_k = dict()
        for k in ks:
            corrupted_heads_hits_at_k[k] = len([x for x in ranks_corrupted_heads if x <= k]) / num_examples
            corrupted_tails_hits_at_k[k] = len([x for x in ranks_corrupted_tails if x <= k]) / num_examples
        return corrupted_heads_hits_at_k, corrupted_tails_hits_at_k

    @staticmethod
    def calculate_mrr(ranks_corrupted_heads, ranks_corrupted_tails, num_examples):
        inverse_ranks_corrupted_heads = [1 / n for n in ranks_corrupted_heads]
        inverse_ranks_corrupted_tails = [1 / n for n in ranks_corrupted_tails]
        mrr_heads = sum(inverse_ranks_corrupted_heads) / num_examples
        mrr_tails = sum(inverse_ranks_corrupted_tails) / num_examples
        return mrr_heads, mrr_tails

    @staticmethod
    def calculate_roc_curve(labels, scores):
        from sklearn.metrics import roc_curve
        fpr, tpr, _ = roc_curve(labels, scores)
        return list(fpr), list(tpr)

    @staticmethod
    def calculate_pr_curve(labels, scores):
        from sklearn.metrics import precision_recall_curve
        precision, recall, thresholds = precision_recall_curve(labels, scores)
        return list(precision), list(recall)

    @staticmethod
    def calculate_auc(x_values, y_values):
        from sklearn.metrics import auc
        auc_value = auc(x_values, y_values)
        return auc_value



from openbiolink.graph_creation.file_reader.mapping.mapDisGeNetReader import MapDisGeNetReader
from openbiolink.graph_creation.file_reader.mapping.mapDrugCentralPubchemReader import MapDrugCentralPubchemReader
from openbiolink.graph_creation.file_reader.mapping.mapStringReader import MapStringReader
from openbiolink.graph_creation.file_reader.mapping.mapUniprotReader import MapUniprotReader



from openbiolink.graph_creation.metadata_db_file.mapping.dbMetadataMapping import DbMetadataMapping
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaMapDisGeNet(DbMetadataMapping):
    NAME = 'Mapping - DisGeNet - Disease (umls -> do)'
    # URL = "http://www.disgenet.org/ds/DisGeNET/results/disease_mappings.tsv.gz"
    URL = "http://www.disgenet.org/static/disgenet_ap1/files/downloads/disease_mappings.tsv.gz"
    OFILE_NAME = "DisGeNet_mapping_disease_umls_do.tab.gz"
    COLS = ['umlsID', 'name', 'voc', 'code', 'vocName']
    FILTER_COLS = ['umlsID', 'voc', 'code']
    HEADER = 1
    DB_TYPE = DbType.DB_MAP_DISGENET

    def __init__(self):
        super().__init__(url=DbMetaMapDisGeNet.URL,
                         ofile_name=DbMetaMapDisGeNet.OFILE_NAME,
                         dbType=DbMetaMapDisGeNet.DB_TYPE)



TTS_FOLDER_NAME = "train_test_data"
CROSS_VAL_FOLDER_NAME = "cross_val"
FOLD_FOLDER_PREFIX = "fold_"

TEST_FILE_NAME = "test_sample.csv"
TRAIN_FILE_NAME = "train_sample.csv"
VAL_FILE_NAME = "val_sample.csv"

NEW_TEST_NODES_FILE_NAME = "removed_test_nodes.csv"
NEW_VAL_NODES_FILE_NAME = "removed_val_nodes.csv"

TRAIN_VAL_NODES_FILE_NAME = "train_val_nodes.csv"
TEST_NODES_FILE_NAME = "test_nodes.csv"

VANISHED_FILE_NAME = "vanished_edges.csv"

EDGE_TYPE_KEY_NAME = 'edge_type_key'



class EdgeMetadata:
    MAP1_META_CLASS = None
    MAP2_META_CLASS = None
    MAP1_ALT_ID_META_CLASS = None
    MAP2_ALT_ID_META_CLASS = None

    def __init__(self, is_directional, edges_file_path, colindex1, colindex2, edgeType, node1_type, node2_type,
                 colindex_qscore=None, cutoff_num=None, cutoff_txt=None,
                 mapping1_file=None, map1_sourceindex=None, map1_targetindex=None,
                 altid_mapping1_file=None, altid_map1_sourceindex=None, altid_map1_targetindex=None,
                 mapping2_file=None, map2_sourceindex=None, map2_targetindex=None,
                 altid_mapping2_file=None, altid_map2_sourceindex=None, altid_map2_targetindex=None
                 ):
        self.is_directional = is_directional
        self.edges_file_path = edges_file_path
        self.colindex1 = colindex1
        self.colindex2 = colindex2
        self.edgeType = edgeType
        self.node1_type = node1_type
        self.node2_type = node2_type
        self.colindex_qscore = colindex_qscore
        self.cutoff_num = cutoff_num
        self.cutoff_txt = cutoff_txt

        self.mapping1_file = mapping1_file
        self.map1_sourceindex = map1_sourceindex
        self.map1_targetindex = map1_targetindex

        self.altid_mapping1_file = altid_mapping1_file
        self.altid_map1_sourceindex = altid_map1_sourceindex
        self.altid_map1_targetindex = altid_map1_targetindex

        self.mapping2_file = mapping2_file
        self.map2_sourceindex = map2_sourceindex
        self.map2_targetindex = map2_targetindex

        self.altid_mapping2_file = altid_mapping2_file
        self.altid_map2_sourceindex = altid_map2_sourceindex
        self.altid_map2_targetindex = altid_map2_targetindex



WORKING_DIR = ''
LOG_FILE_PATH = ''
GUI_MODE = False
RANDOM_STATE = 42  # do not change for reproducibility
GRAPH_PROP_FILE_NAME = 'graph_properties.json'

INTERACTIVE_MODE = True
SKIP_EXISTING_FILES = True

# graph_files col names
NODE1_ID_COL_NAME = 'id1'
NODE2_ID_COL_NAME = 'id2'
EDGE_TYPE_COL_NAME = 'edgeType'
QSCORE_COL_NAME = 'qscore'
VALUE_COL_NAME = 'value'
SCORE_COL_NAME = 'score'
COL_NAMES_TRIPLES = [NODE1_ID_COL_NAME, EDGE_TYPE_COL_NAME, NODE2_ID_COL_NAME]
COL_NAMES_EDGES = [NODE1_ID_COL_NAME, EDGE_TYPE_COL_NAME, NODE2_ID_COL_NAME, QSCORE_COL_NAME]
COL_NAMES_SAMPLES = [NODE1_ID_COL_NAME, EDGE_TYPE_COL_NAME, NODE2_ID_COL_NAME, QSCORE_COL_NAME, VALUE_COL_NAME]

# node files col names
ID_NODE_COL_NAME = 'id'
NODE_TYPE_COL_NAME = 'nodeType'
COL_NAMES_NODES = [ID_NODE_COL_NAME, NODE_TYPE_COL_NAME]



import numpy
import pandas
from tqdm import tqdm

import openbiolink.train_test_set_creation.ttsConfig as ttsConst
from openbiolink import globalConfig as globConst
from openbiolink import utils


class Sampler():

    def __init__(self, meta_edges_dic, nodes):
        self.meta_edges_dic = meta_edges_dic
        self.nodes = nodes

    def generate_n_random_samples(self, n, node_type_1, edge_type, node_type_2, exclude_df):
        exclude_df = exclude_df[globConst.COL_NAMES_EDGES]
        samples = pandas.DataFrame(columns=globConst.COL_NAMES_EDGES)
        nodes_nodeType1 = self.nodes.loc[self.nodes[globConst.NODE_TYPE_COL_NAME] == node_type_1]
        num_nodes1, _ = nodes_nodeType1.shape
        nodes_nodeType2 = self.nodes.loc[self.nodes[globConst.NODE_TYPE_COL_NAME] == node_type_2]
        num_nodes2, _ = nodes_nodeType2.shape
        i = 0
        while len(samples) < n:
            if i > 100:
                break
            num_examples = n - len(samples)
            node1_list = nodes_nodeType1.sample(n=num_examples, random_state=(globConst.RANDOM_STATE + i), replace=True)
            node1_list = node1_list.id.tolist()
            node2_list = nodes_nodeType2.sample(n=num_examples, random_state=(globConst.RANDOM_STATE + (i + 100)),
                                                replace=True)
            node2_list = node2_list.id.tolist()
            sample_candidates = pandas.DataFrame(data={globConst.NODE1_ID_COL_NAME: node1_list,
                                                       globConst.EDGE_TYPE_COL_NAME: [edge_type] * num_examples,
                                                       globConst.NODE2_ID_COL_NAME: node2_list,
                                                       })
            _, sub_samples = utils.get_diff(exclude_df[globConst.COL_NAMES_TRIPLES], sample_candidates)
            sub_samples.drop_duplicates(inplace=True)
            sub_samples[globConst.QSCORE_COL_NAME] = [None] * len(sub_samples)
            samples = samples.append(sub_samples, ignore_index=True)
            exclude_df = exclude_df.append(pandas.DataFrame(sub_samples))
            i += 1
            # todo auswirkungen if num neg examples != num pos examples
        return samples


class NegativeSampler(Sampler):
    def __init__(self, meta_edges_dic, tn_edgeTypes, all_tn, nodes):
        super().__init__(meta_edges_dic, nodes)
        self.meta_edges_dic = meta_edges_dic
        self.tn_edgeTypes = tn_edgeTypes
        self.all_tn = all_tn
        self.all_tn = self.add_edge_type_key_column(all_tn)

    def add_edge_type_key_column(self, df):
        df[ttsConst.EDGE_TYPE_KEY_NAME] = df[globConst.NODE1_ID_COL_NAME].str.split('_').map(lambda x: x[0]) \
                                          + '_' + df[globConst.EDGE_TYPE_COL_NAME] + '_' \
                                          + df[globConst.NODE2_ID_COL_NAME].str.split('_').map(lambda x: x[0])
        return df

    def generate_random_neg_samples(self, pos_samples, distrib='orig'):
        col_names = globConst.COL_NAMES_EDGES
        pos_samples = pos_samples[col_names]
        neg_samples = pandas.DataFrame(columns=col_names)
        pos_samples = self.add_edge_type_key_column(pos_samples)

        # generate distribution of meta_edge types for negative samples
        meta_edges = list(self.meta_edges_dic.keys())
        meta_edges.sort()
        neg_samples_count_meta_edges = {}
        if distrib == 'uni':
            num_tp_examples, _ = pos_samples.shape
            neg_samples_metaEdges = (list(numpy.random.choice(meta_edges, num_tp_examples)))
            neg_samples_metaEdges.sort()
            neg_samples_count_meta_edges = {e: neg_samples_metaEdges.count(e) for e in
                                            set(neg_samples_metaEdges) if neg_samples_metaEdges.count(e) > 0}
        elif distrib == 'orig':
            for key in self.meta_edges_dic.keys():
                num_entry = len(pos_samples.loc[(pos_samples[ttsConst.EDGE_TYPE_KEY_NAME] == key)])
                if num_entry > 0:
                    neg_samples_count_meta_edges[key] = num_entry

        # generate a negative sub-sample for each negative meta_edge type
        for meta_edge_triple_key, count in tqdm(sorted(neg_samples_count_meta_edges.items())):
            node_type_1, edge_type, node_type_2 = self.meta_edges_dic[meta_edge_triple_key]
            pos_samples_of_meta_edge = pos_samples.loc[
                (pos_samples[ttsConst.EDGE_TYPE_KEY_NAME] == meta_edge_triple_key)]

            if edge_type in self.tn_edgeTypes:  # only onto edgesTypes can appear multiple times, there should be no onto tn
                neg_samples = neg_samples.append(self.subsample_with_tn(meta_edge_triple_key=meta_edge_triple_key,
                                                                        subsample_size=count,
                                                                        exclude_df=pos_samples_of_meta_edge[col_names])
                                                 , ignore_index=True)
            else:
                neg_samples = neg_samples.append(self.generate_n_random_samples(n=count,
                                                                                node_type_1=node_type_1,
                                                                                edge_type=edge_type,
                                                                                node_type_2=node_type_2,
                                                                                exclude_df=pos_samples_of_meta_edge[
                                                                                    col_names])
                                                 , ignore_index=True)
        neg_samples[globConst.VALUE_COL_NAME] = 0

        return neg_samples[col_names + [globConst.VALUE_COL_NAME]]

    def subsample_with_tn(self, meta_edge_triple_key, subsample_size, exclude_df, col_names=globConst.COL_NAMES_EDGES):
        node_type_1, edge_type, node_type_2 = self.meta_edges_dic[meta_edge_triple_key]
        tn_examples = self.all_tn.loc[self.all_tn[ttsConst.EDGE_TYPE_KEY_NAME] == meta_edge_triple_key]  # testme
        count_existing_tn, _ = tn_examples.shape
        if subsample_size <= count_existing_tn:
            neg_samples = tn_examples.sample(n=subsample_size, random_state=globConst.RANDOM_STATE)
        else:
            exclude_df = exclude_df.append(tn_examples)
            neg_samples = tn_examples
            neg_samples = neg_samples.append(self.generate_n_random_samples(n=(subsample_size - count_existing_tn),
                                                                            node_type_1=node_type_1,
                                                                            edge_type=edge_type,
                                                                            node_type_2=node_type_2,
                                                                            exclude_df=exclude_df))
        neg_samples.reset_index(inplace=True)

        return neg_samples[col_names]



from openbiolink.edgeType import EdgeType
from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.nodeType import NodeType


class InMetaEdgeHpoGene(InfileMetadata):
    CSV_NAME = "DB_HPO_gene_phenotype.csv"
    USE_COLS = ['geneID', 'hpoID']
    NODE1_COL = 0
    NODE2_COL = 1
    QSCORE_COL = None
    NODE1_TYPE = NodeType.GENE
    NODE2_TYPE = NodeType.PHENOTYPE
    EDGE_TYPE = EdgeType.GENE_PHENOTYPE
    INFILE_TYPE = InfileType.IN_EDGE_HPO_GENE

    MAPPING_SEP = None

    def __init__(self):
        super().__init__(csv_name=InMetaEdgeHpoGene.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaEdgeHpoGene.INFILE_TYPE)



import pandas


class OboParser(object):

    def obo_to_df(self, content, quadruple_list):
        """each quadruple consists of (1) beginning of the line, (2) split character,
        (3) index of split element being the id, (4) the name of dict entry (col_name)"""

        terms = []
        lines = content.readlines()
        term = {}
        for line in lines:
            if line.startswith('[Typedef]'):
                break
            if line.startswith('[Term]'):
                if term:
                    term = self.dic_list_to_dic_string(term)
                    terms.append(term)
                term = {}
                continue
            for tuple in quadruple_list:
                if line.startswith(tuple[0]):
                    elements = line.split(tuple[1])
                    if tuple[3] not in term:
                        term[tuple[3]] = [elements[tuple[2]].strip()]
                    else:
                        term[tuple[3]].append(elements[tuple[2]].strip())
                    break
        term = self.dic_list_to_dic_string(term)
        terms.append(term)
        content.close()
        df = pandas.DataFrame.from_records(terms)
        return df

    def dic_list_to_dic_string(self, dic):
        for k, v in dic.items():
            dic[k] = self.to_string(v)
        return dic

    def to_string(self, list):
        string = ''
        for e in list:
            if string == '':
                string = e
            else:
                string = string + ';' + e
        return string



import os

from openbiolink.graph_creation import graphCreationConfig as g
from openbiolink.graph_creation.file_reader.oboReader import OboReader
from openbiolink.graph_creation.metadata_db_file.onto.dbMetaOntoDo import DbMetaOntoDo
from openbiolink.graph_creation.types.dbType import DbType
from openbiolink.graph_creation.types.readerType import ReaderType


class OntoDoReader(OboReader):
    DB_META_CLASS = DbMetaOntoDo

    def __init__(self):
        super().__init__(
            in_path=os.path.join(g.O_FILE_PATH, self.DB_META_CLASS.OFILE_NAME),
            quadruple_list=self.DB_META_CLASS.QUADRUPLES,
            readerType=ReaderType.READER_ONTO_DO,
            dbType=DbType.DB_ONTO_DO
        )



import pandas

from openbiolink.graph_creation.file_reader.fileReader import FileReader


class CsvReader(FileReader):
    def __init__(self, in_path, readerType, dbType, sep=None, cols=None, use_cols=None, nr_lines_header=0, dtypes=None):
        super().__init__(in_path, readerType, dbType)
        if sep is None:
            self.sep = CsvReader.get_sep(self.in_path)
        else:
            self.sep = sep
        self.cols = cols
        self.use_cols = use_cols
        self.nr_lines_header = nr_lines_header
        self.dtype = dtypes

    @staticmethod
    def get_sep(in_path):
        path_parts = in_path.split('.')
        if (path_parts[1] == "txt"):
            sep = " "
        elif (path_parts[1] == "tsv" or path_parts[1] == "gaf" or path_parts[1] == "tab"):
            sep = "\t"
        else:
            sep = ","
        return sep

    def read_file(self):
        with FileReader.open_file(self.in_path) as in_file:
            data = pandas.read_csv(in_file, sep=self.sep, names=self.cols, usecols=self.use_cols,
                                   skiprows=self.nr_lines_header, dtype=self.dtype)
            return data



import csv
import logging
import os

from tqdm import tqdm

import openbiolink.graphProperties as graphProp
from openbiolink import globalConfig
from openbiolink import globalConfig as globConst
from openbiolink import utils
from openbiolink.cli import Cli
from openbiolink.edge import Edge
from openbiolink.graph_creation import graphCreationConfig as gcConst
from openbiolink.node import Node


class GraphCreator():

    def __init__(self):
        output_dir = os.path.join(globConst.WORKING_DIR, gcConst.GRAPH_FILES_FOLDER_NAME)
        self.tn_path_no_mappings = os.path.join(output_dir, gcConst.TN_ID_NO_MAPPING_FILE_NAME)
        self.tn_path_stats = os.path.join(output_dir, gcConst.TN_STATS_FILE_NAME)
        self.path_no_mappings = os.path.join(output_dir, gcConst.ID_NO_MAPPING_FILE_NAME)
        self.path_stats = os.path.join(output_dir, gcConst.STATS_FILE_NAME)
        os.makedirs(output_dir, exist_ok=True)
        open(self.tn_path_no_mappings, "w").close()
        with open(self.tn_path_stats, "w", encoding="utf8") as file:
            file.write('Edge Type' + '\t' + \
                       'Node1 Type' + '\t' + \
                       'Node2 Type' + '\t' + \
                       'Nr edges' + '\t' + \
                       'Nr edges no mapping' + '\t' + \
                       'Nr edges below cutoff' + '\t' + \
                       'Edges coverage' + '\t' + \
                       'Duplicated edges' + '\t' + \
                       'Nr edges return direction' + '\t' + \
                       'Nr edges after mapping (final nr)' + '\t' + \
                       'Nr nodes1 no mapping' + '\t' + \
                       'Nr nodes2 no mapping' + '\t' + \
                       'Nr nodes1' + '\t' + \
                       'Nr nodes2' + '\t' + \
                       'nodes1 coverage' + '\t' + \
                       'nodes2 coverage' +
                       '\n')
        open(self.path_no_mappings, 'w').close()
        with open(self.path_stats, "w", encoding="utf8") as file:
            file.write('Edge Type' + '\t' + \
                       'Node1 Type' + '\t' + \
                       'Node2 Type' + '\t' + \
                       'Nr edges' + '\t' + \
                       'Nr edges no mapping' + '\t' + \
                       'Nr edges below cutoff' + '\t' + \
                       'Edges coverage' + '\t' + \
                       'Duplicated edges' + '\t' + \
                       'Nr edges return direction' + '\t' + \
                       'Nr edges after mapping (final nr)' + '\t' + \
                       'Nr nodes1 no mapping' + '\t' + \
                       'Nr nodes2 no mapping' + '\t' + \
                       'Nr nodes1' + '\t' + \
                       'Nr nodes2' + '\t' + \
                       'nodes1 coverage' + '\t' + \
                       'nodes2 coverage' +
                       '\n')

    def meta_edges_to_graph(self, edge_metadata_list, tn=None):
        edges_dic = {}
        nodes_dic = {}
        for d in tqdm(edge_metadata_list):
            nodes1, nodes2, edges = self.create_nodes_and_edges(d, tn)
            if str(d.edgeType) in edges_dic:
                edges_dic[str(d.edgeType)].update(edges)
            else:
                edges_dic[str(d.edgeType)] = edges
            if str(d.node1_type) in nodes_dic:
                nodes_dic[str(d.node1_type)].update(nodes1)
            else:
                nodes_dic[str(d.node1_type)] = nodes1
            if str(d.node2_type) in nodes_dic:
                nodes_dic[str(d.node2_type)].update(nodes2)
            else:
                nodes_dic[str(d.node2_type)] = nodes2
        return nodes_dic, edges_dic

    def create_nodes_and_edges(self, edge_metadata, tn=None):
        if not os.path.isfile(edge_metadata.edges_file_path):
            message = 'File does not exist: %s ! Edgetype %s will not be created' % (
            edge_metadata.edges_file_path, str(edge_metadata.edgeType))
            if globalConfig.INTERACTIVE_MODE:
                if globConst.GUI_MODE:
                    from openbiolink.gui import gui
                    gui.askForExit(message)
                else:
                    Cli.ask_for_exit(message)
            else:
                logging.error(message)
            return set(), set(), set()

        # --- mapping ---
        mapping1 = utils.db_mapping_file_to_dic(edge_metadata.mapping1_file, edge_metadata.map1_sourceindex,
                                                edge_metadata.map1_targetindex)
        mapping2 = utils.db_mapping_file_to_dic(edge_metadata.mapping2_file, edge_metadata.map2_sourceindex,
                                                edge_metadata.map2_targetindex)
        altid_mapping1 = utils.db_mapping_file_to_dic(edge_metadata.altid_mapping1_file,
                                                      edge_metadata.altid_map1_sourceindex,
                                                      edge_metadata.altid_map1_targetindex)
        altid_mapping2 = utils.db_mapping_file_to_dic(edge_metadata.altid_mapping2_file,
                                                      edge_metadata.altid_map2_sourceindex,
                                                      edge_metadata.altid_map2_targetindex)

        for mapping in [edge_metadata.mapping1_file, edge_metadata.mapping2_file, edge_metadata.altid_mapping1_file,
                        edge_metadata.altid_mapping2_file]:
            if mapping is not None:
                infile_folder = os.path.join(globConst.WORKING_DIR, gcConst.IN_FILE_FOLDER_NAME)
                mapping_path = os.path.join(infile_folder, mapping)
                if not os.path.isfile(mapping_path):
                    message = 'File does not exist: %s ! Edgetype %s will not be created' % (
                        edge_metadata.edges_file_path, str(edge_metadata.edgeType))
                    if globalConfig.INTERACTIVE_MODE:
                        if globConst.GUI_MODE:
                            from openbiolink.gui import gui
                            gui.askForExit(message)
                        else:
                            Cli.ask_for_exit(message)
                    else:
                        logging.error(message)
                    return set(), set(), set()

        # --- edges ---
        nodes1 = set()
        nodes2 = set()
        edges = set()
        ids1_no_mapping = set()
        ids2_no_mapping = set()
        ids1 = set()
        ids2 = set()
        nr_edges = 0
        nr_edges_return_dir = 0
        nr_edges_after_mapping = 0
        nr_edges_incl_dup = 0
        nr_edges_below_cutoff = 0
        nr_edges_no_mapping = 0

        no_cutoff_defined = edge_metadata.cutoff_num is None and edge_metadata.cutoff_txt is None

        with open(edge_metadata.edges_file_path, "r", encoding="utf8") as edge_content:

            reader = csv.reader(edge_content, delimiter=";")

            for row in reader:
                raw_id1 = row[edge_metadata.colindex1]
                raw_id2 = row[edge_metadata.colindex2]
                if edge_metadata.colindex_qscore is not None:
                    qscore = row[edge_metadata.colindex_qscore]
                else:
                    qscore = None
                edge_id1 = None
                edge_id2 = None
                ids1.add(raw_id1)
                ids2.add(raw_id2)

                # apply mapping
                if (edge_metadata.mapping1_file is not None and raw_id1 in mapping1):
                    edge_id1 = mapping1.get(raw_id1)
                elif (edge_metadata.mapping1_file is None):
                    edge_id1 = [raw_id1]
                if (edge_metadata.mapping2_file is not None and raw_id2 in mapping2):
                    edge_id2 = mapping2.get(raw_id2)
                elif (edge_metadata.mapping2_file is None):
                    edge_id2 = [raw_id2]

                # if mapped successfully
                if edge_id1 is not None and edge_id2 is not None:
                    for id1 in edge_id1:
                        # apply alt_id mapping 1
                        if (edge_metadata.altid_mapping1_file is not None and id1 in altid_mapping1):
                            id1 = altid_mapping1[id1][0]  # there should only be one
                        for id2 in edge_id2:
                            # apply alt_id mapping 2
                            if (edge_metadata.altid_mapping2_file is not None and id2 in altid_mapping2):
                                id2 = altid_mapping2[id2][0]  # there should only be one
                            # check for quality cutoff
                            within_num_cutoff = edge_metadata.cutoff_num is not None and float(
                                qscore) > edge_metadata.cutoff_num
                            within_text_cutoff = edge_metadata.cutoff_txt is not None and qscore not in edge_metadata.cutoff_txt
                            if no_cutoff_defined or within_num_cutoff or within_text_cutoff:
                                bimeg_id1 = edge_metadata.node1_type.name + '_' + id1
                                bimeg_id2 = edge_metadata.node2_type.name + '_' + id2
                                edges.add(Edge(bimeg_id1, edge_metadata.edgeType, bimeg_id2, None, qscore))
                                # add an edge in the other direction when edge is undirectional and graph is directional
                                if (not edge_metadata.is_directional) and graphProp.DIRECTED:
                                    edges.add(Edge(bimeg_id2, edge_metadata.edgeType, bimeg_id1, None, qscore))
                                    nr_edges_incl_dup += 1
                                    nr_edges_return_dir += 1
                                nodes1.add(Node(bimeg_id1, edge_metadata.node1_type))
                                nodes2.add(Node(bimeg_id2, edge_metadata.node2_type))

                                nr_edges_incl_dup += 1
                            else:
                                nr_edges_below_cutoff += 1

                # if not mapped successfully
                else:
                    nr_edges_no_mapping += 1
                    if (edge_id1 is None and edge_metadata.mapping1_file is not None):
                        ids1_no_mapping.add(raw_id1)
                    if (edge_id2 is None and edge_metadata.mapping2_file is not None):
                        ids2_no_mapping.add(raw_id2)
                nr_edges += 1

        nr_edges_after_mapping = len(edges)

        if not no_cutoff_defined and nr_edges_below_cutoff == 0:
            logging.warning(
                "No edges of type %s were cut off by quality cutoff, maybe the metric has changed?" % edge_metadata.edgeType.name)
        if nr_edges_after_mapping == 0:
            logging.warning("No edges of type %s are left after mapping and cutoff!" % edge_metadata.edgeType.name)

        # print statistics
        stats_dic = {
            'edge_type': edge_metadata.edgeType,
            'node1_type': edge_metadata.node1_type,
            'node2_type': edge_metadata.node2_type,
            'nr_edges': nr_edges,
            'nr_edges_below_cutoff': nr_edges_below_cutoff,
            'nr_edges_no_mapping': nr_edges_no_mapping,
            'nr_edges_after_mapping': nr_edges_after_mapping,
            'nr_edges_incl_dup': nr_edges_incl_dup,
            'nr_edges_return_dir': nr_edges_return_dir,
            'ids1_no_mapping': ids1_no_mapping,
            'ids2_no_mapping': ids2_no_mapping,
            'ids1': ids1,
            'ids2': ids2
        }
        self.print_graph_stats(stats_dic, tn)

        return nodes1, nodes2, edges

    def print_graph_stats(self, stats_dic, tn):
        edgeType = stats_dic['edge_type']
        if tn:
            path_no_mappings = self.tn_path_no_mappings
            path_stats = self.tn_path_stats
        else:
            path_no_mappings = self.path_no_mappings
            path_stats = self.path_stats
        with open(path_no_mappings, 'a') as out_file:
            for id in stats_dic['ids1_no_mapping']:
                out_file.write('%s\t%s\n' % (id, edgeType))
            for id in stats_dic['ids2_no_mapping']:
                out_file.write('%s\t%s\n' % (id, edgeType))
            out_file.close()

        stats_string = '\nEdge Type: ' + str(edgeType) + '\n' + \
                       'Node1 Type: ' + str(stats_dic['node1_type']) + '\n' + \
                       'Node2 Type: ' + str(stats_dic['node2_type']) + '\n' + \
                       'Nr edges: ' + str(stats_dic['nr_edges']) + '\n' + \
                       'Nr edges no mapping: ' + str(stats_dic['nr_edges_no_mapping']) + '\n' + \
                       'Nr edges below cutoff: ' + str(stats_dic['nr_edges_below_cutoff']) + '\n' + \
                       'Edges coverage: ' + str(1 - (stats_dic['nr_edges_no_mapping'] / stats_dic['nr_edges'])) + '\n' + \
                       'Duplicated edges: ' + str(
            stats_dic['nr_edges_incl_dup'] - stats_dic['nr_edges_after_mapping']) + '\n' + \
                       'Nr edges after mapping (final nr): ' + str(stats_dic['nr_edges_after_mapping']) + '\n' + \
                       'Nr nodes1 no mapping: ' + str(len(stats_dic['ids1_no_mapping'])) + '\n' + \
                       'Nr nodes2 no mapping: ' + str(len(stats_dic['ids2_no_mapping'])) + '\n' + \
                       'Nr nodes1: ' + str(len(stats_dic['ids1'])) + '\n' + \
                       'Nr nodes2: ' + str(len(stats_dic['ids2'])) + '\n' + \
                       'nodes1 coverage: ' + str(
            1 - (len(stats_dic['ids1_no_mapping']) / len(stats_dic['ids1']))) + '\n' + \
                       'nodes2 coverage: ' + str(
            1 - (len(stats_dic['ids2_no_mapping']) / len(stats_dic['ids2']))) + '\n' + \
                       '######################################################################################\n'

        stats_string = str(edgeType) + '\t' + \
                       str(stats_dic['node1_type']) + '\t' + \
                       str(stats_dic['node2_type']) + '\t' + \
                       str(stats_dic['nr_edges']) + '\t' + \
                       str(stats_dic['nr_edges_no_mapping']) + '\t' + \
                       str(stats_dic['nr_edges_below_cutoff']) + '\t' + \
                       str(1 - (stats_dic['nr_edges_no_mapping'] / stats_dic['nr_edges'])) + '\t' + \
                       str(stats_dic['nr_edges_incl_dup'] - stats_dic['nr_edges_after_mapping']) + '\t' + \
                       str(stats_dic['nr_edges_return_dir']) + '\t' + \
                       str(stats_dic['nr_edges_after_mapping']) + '\t' + \
                       str(len(stats_dic['ids1_no_mapping'])) + '\t' + \
                       str(len(stats_dic['ids2_no_mapping'])) + '\t' + \
                       str(len(stats_dic['ids1'])) + '\t' + \
                       str(len(stats_dic['ids2'])) + '\t' + \
                       str(1 - (len(stats_dic['ids1_no_mapping']) / len(stats_dic['ids1']))) + '\t' + \
                       str(1 - (len(stats_dic['ids2_no_mapping']) / len(stats_dic['ids2']))) + \
                       '\n'
        with open(path_stats, 'a') as out_file:
            out_file.write(stats_string)



from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDisDrug import EdgeMetaDisDrug
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDisPheno import EdgeMetaDisPheno
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugActivationGene import EdgeMetaDrugActivationGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugBindActGene import EdgeMetaDrugBindActGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugBindInhGene import EdgeMetaDrugBindInhGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugBindingGene import EdgeMetaDrugBindingGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugCatalysisGene import EdgeMetaDrugCatalysisGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugExpressionGene import EdgeMetaDrugExpressionGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugInhibitionGene import EdgeMetaDrugInhibitionGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugPheno import EdgeMetaDrugPheno
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugPredBindGene import EdgeMetaDrugPredBindGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaDrugReactionGene import EdgeMetaDrugReactionGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneActivationGene import EdgeMetaGeneActivationGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneAna import EdgeMetaGeneAna
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneBindActGene import EdgeMetaGeneBindActGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneBindInhGene import EdgeMetaGeneBindInhGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneBindingGene import EdgeMetaGeneBindingGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneCatalysisGene import EdgeMetaGeneCatalysisGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneDis import EdgeMetaGeneDis
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneDrug import EdgeMetaGeneDrug
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneExpressionGene import EdgeMetaGeneExpressionGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneGene import EdgeMetaGeneGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneGo import EdgeMetaGeneGo
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneInhibitionGene import EdgeMetaGeneInhibitionGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneOverAna import EdgeMetaGeneOverAna
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGenePath import EdgeMetaGenePath
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGenePheno import EdgeMetaGenePheno
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGenePtmodGene import EdgeMetaGenePtmodGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneReactionGene import EdgeMetaGeneReactionGene
from openbiolink.graph_creation.metadata_edge.edge.edgeMetaGeneUnderAna import EdgeMetaGeneUnderAna
from openbiolink.graph_creation.metadata_edge.edge.tnEdgeMetaDisDrug import TnEdgeMetaDisDrug
from openbiolink.graph_creation.metadata_edge.edge.tnEdgeMetaDisPheno import TnEdgeMetaDisPheno
from openbiolink.graph_creation.metadata_edge.edge.tnEdgeMetaGeneAna import TnEdgeMetaGeneAna
from openbiolink.graph_creation.metadata_edge.edge.tnEdgeMetaGeneOverAna import TnEdgeMetaGeneOverAna
from openbiolink.graph_creation.metadata_edge.edge.tnEdgeMetaGeneUnderAna import TnEdgeMetaGeneUnderAna



from openbiolink.graph_creation.metadata_db_file.mapping.dbMetadataMapping import DbMetadataMapping
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaMapString(DbMetadataMapping):
    NAME = 'Mapping - DisGeNet - Gene (string -> ncbi entrez)'
    # URL = "http://string-db.org/mapping_files/entrez_mappings//entrez_gene_id.vs.string.v10.28042015.tsv"
    URL = "http://string-db.org/mapping_files/entrez/human.entrez_2_string.2018.tsv.gz"
    OFILE_NAME = "String_mapping_gene_ncbi_string.tsv.gz"
    COLS = ['tax_id', 'ncbiID', 'stringID']
    FILTER_COLS = ['ncbiID', 'stringID']
    HEADER = 1
    DB_TYPE = DbType.DB_MAP_STRING

    def __init__(self):
        super().__init__(url=DbMetaMapString.URL,
                         ofile_name=DbMetaMapString.OFILE_NAME,
                         dbType=DbMetaMapString.DB_TYPE)



from openbiolink.nodeType import NodeType


class Node:
    def __init__(self, id: str, type: NodeType, name=""):
        self.id = id
        self.type = type

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.id == other.id
        return False

    def __hash__(self):
        return hash((self.id, self.type))

    def __iter__(self):
        return iter([self.id, self.type])



from openbiolink.graph_creation.metadata_edge.edgeMetadata import EdgeMetadata


class EdgeOntoMetadata(EdgeMetadata):

    def __init__(self, is_directional, edges_file_path, colindex1, colindex2, edgeType, node1_type, node2_type):
        super().__init__(is_directional=is_directional, edges_file_path=edges_file_path,
                         colindex1=colindex1, colindex2=colindex2, edgeType=edgeType, node1_type=node1_type,
                         node2_type=node2_type)



# """ source: #https://github.com/beenje/tkinter-logging-text-widget/blob/master/main.py """
import logging
import queue
from tkinter import ttk, N, S, E, W
from tkinter.scrolledtext import ScrolledText

from openbiolink.gui.gui import *

logger = logging.getLogger()


class QueueHandler(logging.Handler):
    """Class to send logging records to a queue

    It can be used from different threads
    The ConsoleUi class polls this queue to display records in a ScrolledText widget
    """

    # Example from Moshe Kaplan: https://gist.github.com/moshekaplan/c425f861de7bbf28ef06
    # (https://stackoverflow.com/questions/13318742/python-logging-to-tkinter-text-widget) is not thread safe!
    # See https://stackoverflow.com/questions/43909849/tkinter-python-crashes-on-new-thread-trying-to-log-on-main-thread

    def __init__(self, log_queue):
        super().__init__()
        self.log_queue = log_queue

    def emit(self, record):
        self.log_queue.put(record)


class ConsoleUi:
    """Poll messages from a logging queue and display them in a scrolled text widget"""

    def __init__(self, frame):
        self.frame = frame
        # Create a ScrolledText wdiget
        self.scrolled_text = ScrolledText(frame, state='disabled', height=12)
        self.scrolled_text.grid(row=0, column=0, sticky=(N, S, W, E))
        self.scrolled_text.configure(font='TkFixedFont')
        self.scrolled_text.tag_config('INFO', foreground='black')
        self.scrolled_text.tag_config('DEBUG', foreground='gray')
        self.scrolled_text.tag_config('WARNING', foreground='orange')
        self.scrolled_text.tag_config('ERROR', foreground='red')
        self.scrolled_text.tag_config('CRITICAL', foreground='red', underline=1)
        # Create a logging handler using a queue
        self.log_queue = queue.Queue()
        self.queue_handler = QueueHandler(self.log_queue)
        formatter = logging.Formatter('%(asctime)s: %(message)s')
        self.queue_handler.setFormatter(formatter)
        logger.addHandler(self.queue_handler)
        # Start polling messages from the queue
        self.frame.after(100, self.poll_log_queue)

    def display(self, record):
        msg = self.queue_handler.format(record)
        self.scrolled_text.configure(state='normal')
        self.scrolled_text.insert(tk.END, msg + '\n', record.levelname)
        self.scrolled_text.configure(state='disabled')
        # Autoscroll to the bottom
        self.scrolled_text.yview(tk.END)

    def poll_log_queue(self):
        # Check every 100ms if there is a new message in the queue to display
        while True:
            try:
                record = self.log_queue.get(block=False)
            except queue.Empty:
                break
            else:
                self.display(record)
        self.frame.after(100, self.poll_log_queue)


class ConsoleFrame(tk.Frame):
    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller

        # Create the panes and frames
        console_frame = ttk.Labelframe(self, text="Console")
        buttons_panel = tk.Frame(self)
        next_button = tk.Button(buttons_panel, text="Cancel", command=lambda: on_closing(), height=1, width=15)

        # Initialize all frames
        self.console = ConsoleUi(console_frame)
        console_frame.pack(side='top', fill='both', expand=True)

        ttk.Separator(self, orient='horizontal').pack(side='top', fill='x', pady=(15, 0), padx=10, anchor='s')
        buttons_panel.pack(side='bottom', padx=15, fill='x')
        next_button.pack(side='left', anchor='w', pady=(5, 10))

    # def quit(self, *args):
    #    self.root.destroy()



import logging

import pandas


class FileProcessor():

    def __init__(self, use_cols, readerType, infileType, mapping_sep=None):
        self.use_cols = use_cols
        self.readerType = readerType
        self.infileType = infileType
        self.mapping_sep = mapping_sep

    def flat_df(self, data):
        """creates a 'flat' df, i.e. no NAN columns and one relationship per row (a -> b,c) becomes (a -> b ; a -> c) and (a,b -> c) becomes (a->c; b->c)"""
        drop_list = sorted(set(list(data)) - set(self.use_cols))
        data = data.drop(drop_list, axis=1)
        data = data.dropna()
        if self.mapping_sep is not None and not data.empty:
            temp = data[data[self.use_cols[0]].str.contains(self.mapping_sep, regex=False)]
            for row in temp.itertuples():
                for alt in row[1].split(self.mapping_sep):
                    data = data.append(pandas.DataFrame([[alt.lstrip(), row[2]]], columns=self.use_cols))
            data = data[~data[self.use_cols[0]].str.contains(self.mapping_sep, regex=False)]
            temp = data[data[self.use_cols[1]].str.contains(self.mapping_sep, regex=False)]
            for row in temp.itertuples():
                for alt in row[2].split(self.mapping_sep):
                    data = data.append(pandas.DataFrame([[row[1], alt.lstrip()]], columns=self.use_cols))
            data = data[~data[self.use_cols[1]].str.contains(self.mapping_sep, regex=False)]
        return data

    def individual_preprocessing(self, data):
        return data

    def individual_postprocessing(self, data):
        return data

    def process(self, data):
        data = self.individual_preprocessing(data)
        data = data[self.use_cols]
        if self.mapping_sep is not None:
            data = self.flat_df(data)
        data = self.individual_postprocessing(data)
        rows, _ = data.shape
        if rows == 0:
            logging.warning('WARNING: The preprocessing of this processor removed all data.')
        return data

    @staticmethod
    def stitch_to_pubchem_id(data, id_col):
        data.iloc[:, id_col] = data[data.columns[id_col]].str[4:].astype(int)
        return data



from openbiolink.graph_creation.metadata_db_file.dbMetadata import DbMetadata


class DbMetadataEdge(DbMetadata):

    def __init__(self, url, ofile_name, dbType):
        super().__init__(url, ofile_name, dbType)
